from __future__ import annotations

from collections.abc import Collection
from datetime import date
from decimal import Decimal as D
from typing import TYPE_CHECKING

import pytest

import polars as pl
from polars.exceptions import InvalidOperationError
from polars.testing import assert_frame_equal, assert_series_equal

if TYPE_CHECKING:
    from collections.abc import Iterator

    from polars._typing import PolarsDataType


def test_struct_logical_is_in() -> None:
    df1 = pl.DataFrame(
        {
            "x": pl.date_range(date(2022, 1, 1), date(2022, 1, 7), eager=True),
            "y": [0, 4, 6, 2, 3, 4, 5],
        }
    )
    df2 = pl.DataFrame(
        {
            "x": pl.date_range(date(2022, 1, 3), date(2022, 1, 9), eager=True),
            "y": [6, 2, 3, 4, 5, 0, 1],
        }
    )

    s1 = df1.select(pl.struct(["x", "y"])).to_series()
    s2 = df2.select(pl.struct(["x", "y"])).to_series()
    assert s1.is_in(s2).to_list() == [False, False, True, True, True, True, True]


def test_struct_logical_is_in_nonullpropagate() -> None:
    s = pl.Series([date(2022, 1, 1), date(2022, 1, 2), date(2022, 1, 3), None])
    df1 = pl.DataFrame(
        {
            "x": s,
            "y": [0, 4, 6, None],
        }
    )
    s = pl.Series([date(2022, 2, 1), date(2022, 1, 2), date(2022, 2, 3), None])
    df2 = pl.DataFrame(
        {
            "x": s,
            "y": [6, 4, 3, None],
        }
    )

    # Left has no nulls, right has nulls
    s1 = df1.select(pl.struct(["x", "y"])).to_series()
    s1 = s1.extend_constant(s1[0], 1)
    s2 = df2.select(pl.struct(["x", "y"])).to_series().extend_constant(None, 1)
    assert s1.is_in(s2, nulls_equal=False).to_list() == [
        False,
        True,
        False,
        True,
        False,
    ]
    assert s1.is_in(s2, nulls_equal=True).to_list() == [
        False,
        True,
        False,
        True,
        False,
    ]

    # Left has nulls, right has no nulls
    s1 = df1.select(pl.struct(["x", "y"])).to_series().extend_constant(None, 1)
    s2 = df2.select(pl.struct(["x", "y"])).to_series()
    s2 = s2.extend_constant(s2[0], 1)
    assert s1.is_in(s2, nulls_equal=False).to_list() == [
        False,
        True,
        False,
        True,
        None,
    ]
    assert s1.is_in(s2, nulls_equal=True).to_list() == [
        False,
        True,
        False,
        True,
        False,
    ]

    # Both have nulls
    # {None, None} is a valid element unaffected by the missing parameter.
    s1 = df1.select(pl.struct(["x", "y"])).to_series().extend_constant(None, 1)
    s2 = df2.select(pl.struct(["x", "y"])).to_series().extend_constant(None, 1)
    assert s1.is_in(s2, nulls_equal=False).to_list() == [
        False,
        True,
        False,
        True,
        None,
    ]
    assert s1.is_in(s2, nulls_equal=True).to_list() == [
        False,
        True,
        False,
        True,
        True,
    ]


@pytest.mark.parametrize("nulls_equal", [False, True])
def test_is_in_bool(nulls_equal: bool) -> None:
    vals = [True, None]
    df = pl.DataFrame({"A": [True, False, None]})
    missing_value = True if nulls_equal else None
    assert df.select(pl.col("A").is_in(vals, nulls_equal=nulls_equal)).to_dict(
        as_series=False
    ) == {"A": [True, False, missing_value]}


def test_is_in_bool_11216() -> None:
    s = pl.Series([False]).is_in([False, None])
    expected = pl.Series([True])
    assert_series_equal(s, expected)


@pytest.mark.parametrize("nulls_equal", [False, True])
def test_is_in_empty_list_4559(nulls_equal: bool) -> None:
    assert pl.Series(["a"]).is_in([], nulls_equal=nulls_equal).to_list() == [False]


def test_is_in_empty_list_4639() -> None:
    df = pl.DataFrame({"a": [1, None]})
    empty_list: list[int] = []

    result = df.with_columns([pl.col("a").is_in(empty_list).alias("a_in_list")])
    expected = pl.DataFrame({"a": [1, None], "a_in_list": [False, None]})
    assert_frame_equal(result, expected)


def test_is_in_struct() -> None:
    df = pl.DataFrame(
        {
            "struct_elem": [{"a": 1, "b": 11}, {"a": 1, "b": 90}],
            "struct_list": [
                [{"a": 1, "b": 11}, {"a": 2, "b": 12}, {"a": 3, "b": 13}],
                [{"a": 3, "b": 3}],
            ],
        }
    )

    assert df.filter(pl.col("struct_elem").is_in("struct_list")).to_dict(
        as_series=False
    ) == {
        "struct_elem": [{"a": 1, "b": 11}],
        "struct_list": [[{"a": 1, "b": 11}, {"a": 2, "b": 12}, {"a": 3, "b": 13}]],
    }


def test_is_in_null_prop() -> None:
    assert pl.Series([None], dtype=pl.Float32).is_in(pl.Series([42])).item() is None
    assert pl.Series([{"a": None}, None], dtype=pl.Struct({"a": pl.Float32})).is_in(
        pl.Series([{"a": 42}], dtype=pl.Struct({"a": pl.Float32}))
    ).to_list() == [False, None]

    assert pl.Series([{"a": None}, None], dtype=pl.Struct({"a": pl.Boolean})).is_in(
        pl.Series([{"a": 42}], dtype=pl.Struct({"a": pl.Boolean}))
    ).to_list() == [False, None]


def test_is_in_9070() -> None:
    assert not pl.Series([1]).is_in(pl.Series([1.99])).item()


def test_is_in_float_list_10764() -> None:
    df = pl.DataFrame(
        {
            "lst": [[1.0, 2.0, 3.0, 4.0, 5.0], [3.14, 5.28]],
            "n": [3.0, 2.0],
        }
    )
    assert df.select(pl.col("n").is_in("lst").alias("is_in")).to_dict(
        as_series=False
    ) == {"is_in": [True, False]}


def test_is_in_df() -> None:
    df = pl.DataFrame({"a": [1, 2, 3]})
    assert df.select(pl.col("a").is_in([1, 2]))["a"].to_list() == [True, True, False]


def test_is_in_series() -> None:
    s = pl.Series(["a", "b", "c"])

    out = s.is_in(["a", "b"])
    assert out.to_list() == [True, True, False]

    # Check if empty list is converted to pl.String
    out = s.is_in([])
    assert out.to_list() == [False] * out.len()

    for x_y_z in (["x", "y", "z"], {"x", "y", "z"}):
        out = s.is_in(x_y_z)
        assert out.to_list() == [False, False, False]

    df = pl.DataFrame({"a": [1.0, 2.0], "b": [1, 4], "c": ["e", "d"]})
    assert df.select(pl.col("a").is_in(pl.col("b"))).to_series().to_list() == [
        True,
        False,
    ]
    assert df.select(pl.col("b").is_in([])).to_series().to_list() == [False] * df.height

    with pytest.raises(
        InvalidOperationError,
        match=r"'is_in' cannot check for List\(String\) values in Int64 data",
    ):
        df.select(pl.col("b").is_in(["x", "x"]))

    # check we don't shallow-copy and accidentally modify 'a' (see: #10072)
    a = pl.Series("a", [1, 2])
    b = pl.Series("b", [1, 3]).is_in(a)

    assert a.name == "a"
    assert_series_equal(b, pl.Series("b", [True, False]))


@pytest.mark.parametrize("nulls_equal", [False, True])
def test_is_in_null(nulls_equal: bool) -> None:
    # No nulls in right
    s = pl.Series([None, None], dtype=pl.Null)
    result = s.is_in([1, 2], nulls_equal=nulls_equal)
    missing_value = False if nulls_equal else None
    expected = pl.Series([missing_value, missing_value], dtype=pl.Boolean)
    assert_series_equal(result, expected)

    # Nulls in right
    s = pl.Series([None, None], dtype=pl.Null)
    result = s.is_in([None, None], nulls_equal=nulls_equal)
    missing_value = True if nulls_equal else None
    expected = pl.Series([missing_value, missing_value], dtype=pl.Boolean)
    assert_series_equal(result, expected)


@pytest.mark.parametrize("nulls_equal", [False, True])
def test_is_in_boolean(nulls_equal: bool) -> None:
    # Nulls in neither left nor right
    s = pl.Series([True, False])
    result = s.is_in([True, False], nulls_equal=nulls_equal)
    expected = pl.Series([True, True])
    assert_series_equal(result, expected)

    # Nulls in left only
    s = pl.Series([True, None])
    result = s.is_in([False, False], nulls_equal=nulls_equal)
    missing_value = False if nulls_equal else None
    expected = pl.Series([False, missing_value])
    assert_series_equal(result, expected)

    # Nulls in right only
    s = pl.Series([True, False])
    result = s.is_in([True, None], nulls_equal=nulls_equal)
    expected = pl.Series([True, False])
    assert_series_equal(result, expected)

    # Nulls in both
    s = pl.Series([True, False, None])
    result = s.is_in([True, None], nulls_equal=nulls_equal)
    missing_value = True if nulls_equal else None
    expected = pl.Series([True, False, missing_value])
    assert_series_equal(result, expected)


@pytest.mark.parametrize("dtype", [pl.List(pl.Boolean), pl.Array(pl.Boolean, 2)])
@pytest.mark.parametrize("nulls_equal", [False, True])
def test_is_in_boolean_list(dtype: PolarsDataType, nulls_equal: bool) -> None:
    # Note list is_in does not propagate nulls.
    df = pl.DataFrame(
        {
            "a": [True, False, None, None, None],
            "b": pl.Series(
                [
                    [True, False],
                    [True, True],
                    [None, True],
                    [False, True],
                    [True, True],
                ],
                dtype=dtype,
            ),
        }
    )
    missing_true = True if nulls_equal else None
    missing_false = False if nulls_equal else None
    result = df.select(pl.col("a").is_in("b", nulls_equal=nulls_equal))["a"]
    expected = pl.Series("a", [True, False, missing_true, missing_false, missing_false])
    assert_series_equal(result, expected)


def test_is_in_invalid_shape() -> None:
    with pytest.raises(InvalidOperationError):
        pl.Series("a", [1, 2, 3]).is_in([[], []])


def test_is_in_list_rhs() -> None:
    assert_series_equal(
        pl.Series([1, 2, 3, 4, 5]).is_in(pl.Series([[1], [2, 9], [None], None, None])),
        pl.Series([True, True, False, None, None]),
    )


@pytest.mark.parametrize("dtype", [pl.Float32, pl.Float64])
def test_is_in_float(dtype: PolarsDataType) -> None:
    s = pl.Series([float("nan"), 0.0], dtype=dtype)
    result = s.is_in([-0.0, -float("nan")])
    expected = pl.Series([True, True], dtype=pl.Boolean)
    assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("df", "matches", "expected_error"),
    [
        (
            pl.DataFrame({"a": [1, 2], "b": [[1.0, 2.5], [3.0, 4.0]]}),
            [True, False],
            None,
        ),
        (
            pl.DataFrame({"a": [2.5, 3.0], "b": [[1, 2], [3, 4]]}),
            [False, True],
            None,
        ),
        (
            pl.DataFrame(
                {"a": [None, None], "b": [[1, 2], [3, 4]]},
                schema_overrides={"a": pl.Null},
            ),
            [None, None],
            None,
        ),
        (
            pl.DataFrame({"a": ["1", "2"], "b": [[1, 2], [3, 4]]}),
            None,
            r"'is_in' cannot check for List\(Int64\) values in String data",
        ),
        (
            pl.DataFrame({"a": [date.today(), None], "b": [[1, 2], [3, 4]]}),
            None,
            r"'is_in' cannot check for List\(Int64\) values in Date data",
        ),
    ],
)
def test_is_in_expr_list_series(
    df: pl.DataFrame, matches: list[bool] | None, expected_error: str | None
) -> None:
    expr_is_in = pl.col("a").is_in(pl.col("b"))
    if matches:
        assert df.select(expr_is_in).to_series().to_list() == matches
    else:
        with pytest.raises(InvalidOperationError, match=expected_error):
            df.select(expr_is_in)


@pytest.mark.parametrize(
    ("df", "matches"),
    [
        (
            pl.DataFrame({"a": [1, None], "b": [[1.0, 2.5, 4.0], [3.0, 4.0, 5.0]]}),
            [True, False],
        ),
        (
            pl.DataFrame({"a": [1, None], "b": [[0.0, 2.5, None], [3.0, 4.0, None]]}),
            [False, True],
        ),
        (
            pl.DataFrame(
                {"a": [None, None], "b": [[1, 2], [3, 4]]},
                schema_overrides={"a": pl.Null},
            ),
            [False, False],
        ),
        (
            pl.DataFrame(
                {"a": [None, None], "b": [[1, 2], [3, None]]},
                schema_overrides={"a": pl.Null},
            ),
            [False, True],
        ),
    ],
)
def test_is_in_expr_list_series_nonullpropagate(
    df: pl.DataFrame, matches: list[bool]
) -> None:
    expr_is_in = pl.col("a").is_in(pl.col("b"), nulls_equal=True)
    assert df.select(expr_is_in).to_series().to_list() == matches


@pytest.mark.parametrize("nulls_equal", [False, True])
def test_is_in_null_series(nulls_equal: bool) -> None:
    df = pl.DataFrame({"a": ["a", "b", None]})
    result = df.select(pl.col("a").is_in([None], nulls_equal=nulls_equal))
    missing_value = True if nulls_equal else None
    expected = pl.DataFrame({"a": [False, False, missing_value]})
    assert_frame_equal(result, expected)


def test_is_in_int_range() -> None:
    r = pl.int_range(0, 3, eager=False)
    out = pl.select(r.is_in([1, 2])).to_series()
    assert out.to_list() == [False, True, True]

    r = pl.int_range(0, 3, eager=True)  # type: ignore[assignment]
    out = r.is_in([1, 2])  # type: ignore[assignment]
    assert out.to_list() == [False, True, True]


def test_is_in_date_range() -> None:
    r = pl.date_range(date(2023, 1, 1), date(2023, 1, 3), eager=False)
    out = pl.select(r.is_in([date(2023, 1, 2), date(2023, 1, 3)])).to_series()
    assert out.to_list() == [False, True, True]

    r = pl.date_range(date(2023, 1, 1), date(2023, 1, 3), eager=True)  # type: ignore[assignment]
    out = r.is_in([date(2023, 1, 2), date(2023, 1, 3)])  # type: ignore[assignment]
    assert out.to_list() == [False, True, True]


@pytest.mark.parametrize("dtype", [pl.Categorical, pl.Enum(["a", "b", "c"])])
@pytest.mark.parametrize("nulls_equal", [False, True])
def test_cat_is_in_series(dtype: pl.DataType, nulls_equal: bool) -> None:
    s = pl.Series(["a", "b", "c", None], dtype=dtype)
    s2 = pl.Series(["b", "c"], dtype=dtype)
    missing_value = False if nulls_equal else None
    expected = pl.Series([False, True, True, missing_value])
    assert_series_equal(s.is_in(s2, nulls_equal=nulls_equal), expected)

    s2_str = s2.cast(pl.String)
    assert_series_equal(s.is_in(s2_str, nulls_equal=nulls_equal), expected)


@pytest.mark.parametrize("nulls_equal", [False, True])
def test_cat_is_in_series_non_existent(nulls_equal: bool) -> None:
    dtype = pl.Categorical
    s = pl.Series(["a", "b", "c", None], dtype=dtype)
    s2 = pl.Series(["a", "d", "e"], dtype=dtype)
    missing_value = False if nulls_equal else None
    expected = pl.Series([True, False, False, missing_value])
    assert_series_equal(s.is_in(s2, nulls_equal=nulls_equal), expected)

    s2_str = s2.cast(pl.String)
    assert_series_equal(s.is_in(s2_str, nulls_equal=nulls_equal), expected)


@pytest.mark.parametrize(
    "nulls_equal",
    [False, True],
)
def test_enum_is_in_series_non_existent(nulls_equal: bool) -> None:
    dtype = pl.Enum(["a", "b", "c"])
    missing_value = False if nulls_equal else None
    s = pl.Series(["a", "b", "c", None], dtype=dtype)
    s2_str = pl.Series(["a", "d", "e"])
    expected = pl.Series([True, False, False, missing_value])

    with pytest.raises(InvalidOperationError):
        s.is_in(s2_str, nulls_equal=nulls_equal)
    with pytest.raises(InvalidOperationError):
        s.is_in(["a", "d", "e"], nulls_equal=nulls_equal)

    out = s.is_in(["a"], nulls_equal=nulls_equal)
    assert_series_equal(out, expected)


@pytest.mark.parametrize("dtype", [pl.Categorical, pl.Enum(["a", "b", "c"])])
@pytest.mark.parametrize("nulls_equal", [False, True])
def test_cat_is_in_with_lit_str(dtype: pl.DataType, nulls_equal: bool) -> None:
    missing_value = False if nulls_equal else None
    s = pl.Series(["a", "b", "c", None], dtype=dtype)
    lit = ["b"]
    expected = pl.Series([False, True, False, missing_value])

    assert_series_equal(s.is_in(lit, nulls_equal=nulls_equal), expected)


@pytest.mark.parametrize("nulls_equal", [False, True])
def test_cat_is_in_with_lit_str_non_existent(nulls_equal: bool) -> None:
    dtype = pl.Categorical()
    missing_value = False if nulls_equal else None
    s = pl.Series(["a", "b", "c", None], dtype=dtype)
    lit = ["d"]
    expected = pl.Series([False, False, False, missing_value])

    assert_series_equal(s.is_in(lit, nulls_equal=nulls_equal), expected)


@pytest.mark.parametrize("dtype", [pl.Categorical, pl.Enum(["a", "b", "c"])])
def test_cat_is_in_with_lit_str_cache_setup(dtype: pl.DataType) -> None:
    # init the global cache
    _ = pl.Series(["c", "b", "a"], dtype=dtype)

    assert_series_equal(pl.Series(["a"], dtype=dtype).is_in(["a"]), pl.Series([True]))
    assert_series_equal(pl.Series(["b"], dtype=dtype).is_in(["b"]), pl.Series([True]))
    assert_series_equal(pl.Series(["c"], dtype=dtype).is_in(["c"]), pl.Series([True]))


def test_is_in_with_wildcard_13809() -> None:
    out = pl.DataFrame({"A": ["B"]}).select(pl.all().is_in(["C"]))
    expected = pl.DataFrame({"A": [False]})
    assert_frame_equal(out, expected)


@pytest.mark.parametrize(
    "dtype",
    [
        pl.Categorical,
        pl.Enum(["a", "b", "c", "d"]),
    ],
)
def test_cat_is_in_from_str(dtype: pl.DataType) -> None:
    s = pl.Series(["c", "c", "b"], dtype=dtype)

    # test local
    assert_series_equal(
        pl.Series(["a", "d", "b"]).is_in(s),
        pl.Series([False, False, True]),
    )


@pytest.mark.parametrize("dtype", [pl.Categorical, pl.Enum(["a", "b", "c", "d"])])
def test_cat_list_is_in_from_cat(dtype: pl.DataType) -> None:
    df = pl.DataFrame(
        [
            (["a", "b"], "c"),
            (["a", "b"], "a"),
            (["a", None], None),
            (["a", "c"], None),
            (["a"], "d"),
        ],
        schema={"li": pl.List(dtype), "x": dtype},
        orient="row",
    )
    res = df.select(pl.col("li").list.contains(pl.col("x")))
    expected_df = pl.DataFrame({"li": [False, True, True, False, False]})
    assert_frame_equal(res, expected_df)


@pytest.mark.parametrize(
    ("val", "expected"),
    [
        ("b", [True, False, False, None, True]),
        (None, [False, False, True, None, False]),
        ("e", [False, False, False, None, False]),
    ],
)
def test_cat_list_is_in_from_cat_single(val: str | None, expected: list[bool]) -> None:
    df = pl.Series(
        "li",
        [["a", "b"], ["a", "c"], ["a", None], None, ["b"]],
        dtype=pl.List(pl.Categorical),
    ).to_frame()
    res = df.select(pl.col("li").list.contains(pl.lit(val, dtype=pl.Categorical)))
    expected_df = pl.DataFrame({"li": expected})
    assert_frame_equal(res, expected_df)


def test_cat_list_is_in_from_str() -> None:
    df = pl.DataFrame(
        [
            (["a", "b"], "c"),
            (["a", "b"], "a"),
            (["a", None], None),
            (["a", "c"], None),
            (["a"], "d"),
        ],
        schema={"li": pl.List(pl.Categorical), "x": pl.String},
        orient="row",
    )
    res = df.select(pl.col("li").list.contains(pl.col("x")))
    expected_df = pl.DataFrame({"li": [False, True, True, False, False]})
    assert_frame_equal(res, expected_df)


@pytest.mark.parametrize(
    ("val", "expected"),
    [
        ("b", [True, False, False, None, True]),
        (None, [False, False, True, None, False]),
        ("e", [False, False, False, None, False]),
    ],
)
def test_cat_list_is_in_from_single_str(val: str | None, expected: list[bool]) -> None:
    df = pl.Series(
        "li",
        [["a", "b"], ["a", "c"], ["a", None], None, ["b"]],
        dtype=pl.List(pl.Categorical),
    ).to_frame()
    res = df.select(pl.col("li").list.contains(pl.lit(val, dtype=pl.String)))
    expected_df = pl.DataFrame({"li": expected})
    assert_frame_equal(res, expected_df)


@pytest.mark.parametrize("nulls_equal", [False, True])
def test_is_in_struct_enum_17618(nulls_equal: bool) -> None:
    df = pl.DataFrame()
    dtype = pl.Enum(categories=["HBS"])
    df = df.insert_column(0, pl.Series("category", [], dtype=dtype))
    assert df.filter(
        pl.struct("category").is_in(
            pl.Series(
                [{"category": "HBS"}],
                dtype=pl.Struct({"category": df["category"].dtype}),
            ),
            nulls_equal=nulls_equal,
        )
    ).shape == (0, 1)


@pytest.mark.parametrize("nulls_equal", [False, True])
def test_is_in_decimal(nulls_equal: bool) -> None:
    assert pl.DataFrame({"a": [D("0.0"), D("0.2"), D("0.1")]}).select(
        pl.col("a").is_in([0.0, 0.1], nulls_equal=nulls_equal)
    )["a"].to_list() == [True, False, True]
    assert pl.DataFrame({"a": [D("0.0"), D("0.2"), D("0.1")]}).select(
        pl.col("a").is_in([D("0.0"), D("0.1")], nulls_equal=nulls_equal)
    )["a"].to_list() == [True, False, True]
    assert pl.DataFrame({"a": [D("0.0"), D("0.2"), D("0.1")]}).select(
        pl.col("a").is_in([1, 0, 2], nulls_equal=nulls_equal)
    )["a"].to_list() == [True, False, False]
    missing_value = True if nulls_equal else None
    assert pl.DataFrame({"a": [D("0.0"), D("0.2"), None]}).select(
        pl.col("a").is_in([0.0, 0.1, None], nulls_equal=nulls_equal)
    )["a"].to_list() == [True, False, missing_value]
    missing_value = False if nulls_equal else None
    assert pl.DataFrame({"a": [D("0.0"), D("0.2"), None]}).select(
        pl.col("a").is_in([0.0, 0.1], nulls_equal=nulls_equal)
    )["a"].to_list() == [True, False, missing_value]


def test_is_in_collection() -> None:
    df = pl.DataFrame(
        {
            "lbl": ["aa", "bb", "cc", "dd", "ee"],
            "val": [0, 1, 2, 3, 4],
        }
    )

    class CustomCollection(Collection[int]):
        def __init__(self, vals: Collection[int]) -> None:
            super().__init__()
            self.vals = vals

        def __contains__(self, x: object) -> bool:
            return x in self.vals

        def __iter__(self) -> Iterator[int]:
            yield from self.vals

        def __len__(self) -> int:
            return len(self.vals)

    for constraint_values in (
        {3, 2, 1},
        frozenset({3, 2, 1}),
        CustomCollection([3, 2, 1]),
    ):
        res = df.filter(pl.col("val").is_in(constraint_values))
        assert set(res["lbl"]) == {"bb", "cc", "dd"}


@pytest.mark.parametrize("nulls_equal", [False, True])
def test_null_propagate_all_paths(nulls_equal: bool) -> None:
    # No nulls in either
    s = pl.Series([1, 2, 3])
    result = s.is_in([1, 3, 8], nulls_equal=nulls_equal)
    expected = pl.Series([True, False, True])
    assert_series_equal(result, expected)

    # Nulls in left only
    s = pl.Series([1, 2, None])
    result = s.is_in([1, 3, 8], nulls_equal=nulls_equal)
    missing_value = False if nulls_equal else None
    expected = pl.Series([True, False, missing_value])
    assert_series_equal(result, expected)

    # Nulls in right only
    s = pl.Series([1, 2, 3])
    result = s.is_in([1, 3, None], nulls_equal=nulls_equal)
    expected = pl.Series([True, False, True])
    assert_series_equal(result, expected)

    # Nulls in both
    s = pl.Series([1, 2, None])
    result = s.is_in([1, 3, None], nulls_equal=nulls_equal)
    missing_value = True if nulls_equal else None
    expected = pl.Series([True, False, missing_value])
    assert_series_equal(result, expected)


@pytest.mark.parametrize("nulls_equal", [False, True])
def test_null_propagate_all_paths_cat(nulls_equal: bool) -> None:
    # No nulls in either
    s = pl.Series(["1", "2", "3"])
    result = s.is_in(["1", "3", "8"], nulls_equal=nulls_equal)
    expected = pl.Series([True, False, True])
    assert_series_equal(result, expected)

    # Nulls in left only
    s = pl.Series(["1", "2", None])
    result = s.is_in(["1", "3", "8"], nulls_equal=nulls_equal)
    missing_value = False if nulls_equal else None
    expected = pl.Series([True, False, missing_value])
    assert_series_equal(result, expected)

    # Nulls in right only
    s = pl.Series(["1", "2", "3"])
    result = s.is_in(["1", "3", None], nulls_equal=nulls_equal)
    expected = pl.Series([True, False, True])
    assert_series_equal(result, expected)

    # Nulls in both
    s = pl.Series(["1", "2", None])
    result = s.is_in(["1", "3", None], nulls_equal=nulls_equal)
    missing_value = True if nulls_equal else None
    expected = pl.Series([True, False, missing_value])
    assert_series_equal(result, expected)



from __future__ import annotations

import sys
import typing

import pytest

import polars as pl
from polars.testing import assert_frame_equal, assert_series_equal
from tests.unit.conftest import INTEGER_DTYPES


@pytest.mark.parametrize("op", ["and_", "or_"])
def test_bitwise_integral_schema(op: str) -> None:
    df = pl.LazyFrame({"a": [1, 2], "b": [3, 4]})
    q = df.select(getattr(pl.col("a"), op)(pl.col("b")))
    assert q.collect_schema()["a"] == df.collect_schema()["a"]


@pytest.mark.parametrize("op", ["and_", "or_", "xor"])
def test_bitwise_single_null_value_schema(op: str) -> None:
    df = pl.DataFrame({"a": [True, True]})
    q = df.select(getattr(pl.col("a"), op)(None))
    result_schema = q.collect_schema()
    assert result_schema.len() == 1
    assert "a" in result_schema


def leading_zeros(v: int | None, nb: int) -> int | None:
    if v is None:
        return None

    b = bin(v)[2:]
    blen = len(b) - len(b.lstrip("0"))
    if blen == len(b):
        return nb
    else:
        return nb - len(b) + blen


def leading_ones(v: int | None, nb: int) -> int | None:
    if v is None:
        return None

    b = bin(v)[2:]
    if len(b) < nb:
        return 0
    else:
        return len(b) - len(b.lstrip("1"))


def trailing_zeros(v: int | None, nb: int) -> int | None:
    if v is None:
        return None

    b = bin(v)[2:]
    blen = len(b) - len(b.rstrip("0"))
    if blen == len(b):
        return nb
    else:
        return blen


def trailing_ones(v: int | None) -> int | None:
    if v is None:
        return None

    b = bin(v)[2:]
    return len(b) - len(b.rstrip("1"))


@pytest.mark.parametrize(
    "value",
    [
        0x00,
        0x01,
        0xFCEF_0123,
        0xFFFF_FFFF,
        0xFFF0_FFE1_ABCD_EF01,
        0xAAAA_AAAA_AAAA_AAAA,
        None,
    ],
)
@pytest.mark.parametrize("dtype", [*INTEGER_DTYPES, pl.Boolean])
@pytest.mark.skipif(sys.version_info < (3, 10), reason="bit_count introduced in 3.10")
@typing.no_type_check
def test_bit_counts(value: int, dtype: pl.DataType) -> None:
    bitsize = 8
    if "Boolean" in str(dtype):
        bitsize = 1
    if "16" in str(dtype):
        bitsize = 16
    elif "32" in str(dtype):
        bitsize = 32
    elif "64" in str(dtype):
        bitsize = 64
    elif "128" in str(dtype):
        bitsize = 128

    if bitsize == 1 and value is not None:
        value = value & 1 != 0

        co = 1 if value else 0
        cz = 0 if value else 1
    elif value is not None:
        value = value & ((1 << bitsize) - 1)

        if dtype.is_signed_integer() and value >> (bitsize - 1) > 0:
            value = value - pow(2, bitsize - 1)

        co = value.bit_count()
        cz = bitsize - co
    else:
        co = None
        cz = None

    assert_series_equal(
        pl.Series("a", [value], dtype).bitwise_count_ones(),
        pl.Series("a", [co], pl.UInt32),
    )
    assert_series_equal(
        pl.Series("a", [value], dtype).bitwise_count_zeros(),
        pl.Series("a", [cz], pl.UInt32),
    )
    assert_series_equal(
        pl.Series("a", [value], dtype).bitwise_leading_ones(),
        pl.Series("a", [leading_ones(value, bitsize)], pl.UInt32),
    )
    assert_series_equal(
        pl.Series("a", [value], dtype).bitwise_leading_zeros(),
        pl.Series("a", [leading_zeros(value, bitsize)], pl.UInt32),
    )
    assert_series_equal(
        pl.Series("a", [value], dtype).bitwise_trailing_ones(),
        pl.Series("a", [trailing_ones(value)], pl.UInt32),
    )
    assert_series_equal(
        pl.Series("a", [value], dtype).bitwise_trailing_zeros(),
        pl.Series("a", [trailing_zeros(value, bitsize)], pl.UInt32),
    )


@pytest.mark.parametrize("dtype", INTEGER_DTYPES)
def test_bit_aggregations(dtype: pl.DataType) -> None:
    s = pl.Series("a", [0x74, 0x1C, 0x05], dtype)

    df = s.to_frame().select(
        AND=pl.col.a.bitwise_and(),
        OR=pl.col.a.bitwise_or(),
        XOR=pl.col.a.bitwise_xor(),
    )

    assert_frame_equal(
        df,
        pl.DataFrame(
            [
                pl.Series("AND", [0x04], dtype),
                pl.Series("OR", [0x7D], dtype),
                pl.Series("XOR", [0x6D], dtype),
            ]
        ),
    )


@pytest.mark.parametrize("dtype", INTEGER_DTYPES)
def test_bit_aggregations_lazy_no_nulls(dtype: pl.DataType) -> None:
    s = pl.Series("a", [0x74, 0x1C, 0x05], dtype)

    lf = s.to_frame().lazy()

    out = lf.select(
        AND=pl.col.a.bitwise_and(),
        OR=pl.col.a.bitwise_or(),
        XOR=pl.col.a.bitwise_xor(),
    ).collect()

    assert_frame_equal(
        out,
        pl.DataFrame(
            [
                pl.Series("AND", [0x04], dtype),
                pl.Series("OR", [0x7D], dtype),
                pl.Series("XOR", [0x6D], dtype),
            ]
        ),
    )


@pytest.mark.parametrize("dtype", INTEGER_DTYPES)
def test_bit_aggregations_lazy_some_nulls(dtype: pl.DataType) -> None:
    s = pl.Series("a", [0x74, None, 0x1C, None, 0x05], dtype)
    out = (
        s.to_frame()
        .lazy()
        .select(
            AND=pl.col.a.bitwise_and(),
            OR=pl.col.a.bitwise_or(),
            XOR=pl.col.a.bitwise_xor(),
        )
        .collect()
    )

    assert_frame_equal(
        out,
        pl.DataFrame(
            [
                pl.Series("AND", [0x04], dtype),
                pl.Series("OR", [0x7D], dtype),
                pl.Series("XOR", [0x6D], dtype),
            ]
        ),
    )


@pytest.mark.parametrize(
    "expr",
    [pl.col("a").bitwise_and(), pl.col("a").bitwise_or(), pl.col("a").bitwise_xor()],
)
def test_bit_aggregations_lazy_all_nulls(expr: pl.Expr) -> None:
    dtype = pl.Int64
    s = pl.Series("a", [None, None, None], dtype)
    out = s.to_frame().lazy().select(OUT=expr).collect()

    assert_frame_equal(
        out,
        pl.DataFrame([pl.Series("OUT", [None], dtype)]),
    )


@pytest.mark.parametrize("dtype", INTEGER_DTYPES)
def test_bit_group_by(dtype: pl.DataType) -> None:
    df = pl.DataFrame(
        [
            pl.Series("g", [4, 1, 1, 2, 3, 2, 4, 4], pl.Int8),
            pl.Series("a", [0x03, 0x74, 0x1C, 0x05, None, 0x70, 0x01, None], dtype),
        ]
    )

    df = df.group_by("g").agg(
        AND=pl.col.a.bitwise_and(),
        OR=pl.col.a.bitwise_or(),
        XOR=pl.col.a.bitwise_xor(),
    )

    assert_frame_equal(
        df,
        pl.DataFrame(
            [
                pl.Series("g", [1, 2, 3, 4], pl.Int8),
                pl.Series("AND", [0x74 & 0x1C, 0x05 & 0x70, None, 0x01], dtype),
                pl.Series("OR", [0x74 | 0x1C, 0x05 | 0x70, None, 0x03], dtype),
                pl.Series("XOR", [0x74 ^ 0x1C, 0x05 ^ 0x70, None, 0x02], dtype),
            ]
        ),
        check_row_order=False,
    )


def test_scalar_bitwise_xor() -> None:
    df = pl.select(
        pl.repeat(pl.lit(0x80, pl.UInt8), i).bitwise_xor().alias(f"l{i}")
        for i in range(5)
    ).transpose()

    assert_series_equal(
        df.to_series(),
        pl.Series("x", [None, 0x80, 0x00, 0x80, 0x00], pl.UInt8),
        check_names=False,
    )


@pytest.mark.parametrize(
    ("expr", "result"),
    [
        (pl.all().bitwise_and(), [True, False, False, True, False, None]),
        (pl.all().bitwise_or(), [True, True, False, True, False, None]),
        (pl.all().bitwise_xor(), [False, True, False, True, False, None]),
    ],
)
def test_bool_bitwise_with_nulls_23314(expr: pl.Expr, result: list[bool]) -> None:
    df = pl.DataFrame(
        {
            "a": [True, True, None],
            "b": [True, False, None],
            "c": [False, False, None],
            "d": [True, None, None],
            "e": [False, None, None],
            "f": [None, None, None],
        },
        schema_overrides={"f": pl.Boolean},
    )
    columns = ["a", "b", "c", "d", "e", "f"]
    out = df.select(expr)
    expected = pl.DataFrame(
        [result], orient="row", schema=columns, schema_overrides={"f": pl.Boolean}
    )
    assert_frame_equal(out, expected)


@pytest.mark.parametrize(
    ("expr", "result"),
    [
        (pl.all().bitwise_and(), [True, False, False, False, False, None]),
        (pl.all().bitwise_or(), [True, True, True, False, True, None]),
        (pl.all().bitwise_xor(), [True, False, True, False, True, None]),
    ],
)
def test_bitwise_boolean(expr: pl.Expr, result: list[bool]) -> None:
    lf = pl.LazyFrame(
        {
            "a": [True, True, True],
            "b": [True, False, True],
            "c": [False, True, False],
            "d": [False, False, False],
            "x": [True, False, None],
            "z": [None, None, None],
        },
        schema_overrides={"z": pl.Boolean},
    )

    columns = ["a", "b", "c", "d", "x", "z"]
    expected = pl.DataFrame(
        [result], orient="row", schema=columns, schema_overrides={"z": pl.Boolean}
    )
    out = lf.select(expr).collect()
    assert_frame_equal(out, expected)


# Although there is no way to deterministically trigger the `evict` path
# in the code, the below test will do so with high likelihood
# POLARS_MAX_THREADS is only honored when tested in isolation, see issue #22070
def test_bitwise_boolean_evict_path(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("POLARS_MAX_THREADS", "1")
    monkeypatch.setenv("POLARS_HOT_TABLE_SIZE", "2")
    n_groups = 100
    group_size_pairs = 10
    group_size = group_size_pairs * 2

    col_a = list(range(group_size)) * n_groups
    col_b = [True, False] * group_size_pairs * n_groups
    df = pl.DataFrame({"a": pl.Series(col_a), "b": pl.Series(col_b)}).sort("a")

    out = (
        df.lazy()
        .group_by("a")
        .agg(
            [
                pl.col("b").bitwise_and().alias("bitwise_and"),
                pl.col("b").bitwise_or().alias("bitwise_or"),
                pl.col("b").bitwise_xor().alias("bitwise_xor"),
            ]
        )
        .sort("a")
        .collect()
    )
    expected = pl.DataFrame(
        {
            "a": list(range(group_size)),
            "bitwise_and": [True, False] * group_size_pairs,
            "bitwise_or": [True, False] * group_size_pairs,
            "bitwise_xor": [n_groups % 2 == 1, False] * group_size_pairs,
        }
    )
    assert_frame_equal(out, expected)



from __future__ import annotations

import contextlib
import datetime
from typing import TYPE_CHECKING, Any, TypedDict

from hypothesis import Phase, given, settings

import polars as pl
from polars.meta import get_index_type
from polars.testing import assert_frame_equal, assert_series_equal
from polars.testing.parametric.strategies import series

if TYPE_CHECKING:
    from collections.abc import Sequence

    from polars._typing import PythonLiteral


class Case(TypedDict):
    """A test case for Skip Batch Predicate."""

    min: Any | None
    max: Any | None
    null_count: int | None
    len: int | None
    can_skip: bool


def assert_skp_series(
    name: str,
    dtype: pl.DataType,
    expr: pl.Expr,
    cases: Sequence[Case],
) -> None:
    sbp = expr._skip_batch_predicate({name: dtype})

    df = pl.DataFrame(
        [
            pl.Series(f"{name}_min", [i["min"] for i in cases], dtype),
            pl.Series(f"{name}_max", [i["max"] for i in cases], dtype),
            pl.Series(f"{name}_nc", [i["null_count"] for i in cases], get_index_type()),
            pl.Series("len", [i["len"] for i in cases], get_index_type()),
        ]
    )
    mask = pl.Series("can_skip", [i["can_skip"] for i in cases], pl.Boolean)

    out = df.select(can_skip=sbp).to_series()
    out = out.replace(None, False)

    try:
        assert_series_equal(out, mask)
    except AssertionError:
        print(sbp)
        raise


def test_true_false_predicate() -> None:
    true_sbp = pl.lit(True)._skip_batch_predicate({})
    false_sbp = pl.lit(False)._skip_batch_predicate({})
    null_sbp = pl.lit(None)._skip_batch_predicate({})

    df = pl.DataFrame({"len": [1]})

    out = df.select(
        true=true_sbp,
        false=false_sbp,
        null=null_sbp,
    )

    assert_frame_equal(
        out,
        pl.DataFrame(
            {
                "true": [False],
                "false": [True],
                "null": [True],
            }
        ),
    )


def test_equality() -> None:
    assert_skp_series(
        "a",
        pl.Int64(),
        pl.col("a") == 5,
        [
            {"min": 1, "max": 2, "null_count": 0, "len": 42, "can_skip": True},
            {"min": 6, "max": 7, "null_count": 0, "len": 42, "can_skip": True},
            {"min": 1, "max": 7, "null_count": 0, "len": 42, "can_skip": False},
            {"min": None, "max": None, "null_count": 42, "len": 42, "can_skip": True},
        ],
    )

    assert_skp_series(
        "a",
        pl.Int64(),
        pl.col("a") != 0,
        [
            {"min": 0, "max": 0, "null_count": 6, "len": 7, "can_skip": False},
        ],
    )


def test_datetimes() -> None:
    d = datetime.datetime(2023, 4, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)
    td = datetime.timedelta

    assert_skp_series(
        "a",
        pl.Datetime(time_zone=datetime.timezone.utc),
        pl.col("a") == d,
        [
            {
                "min": d - td(days=2),
                "max": d - td(days=1),
                "null_count": 0,
                "len": 42,
                "can_skip": True,
            },
            {
                "min": d + td(days=1),
                "max": d - td(days=2),
                "null_count": 0,
                "len": 42,
                "can_skip": True,
            },
            {"min": d, "max": d, "null_count": 42, "len": 42, "can_skip": True},
            {"min": d, "max": d, "null_count": 0, "len": 42, "can_skip": False},
            {
                "min": d - td(days=2),
                "max": d + td(days=2),
                "null_count": 0,
                "len": 42,
                "can_skip": False,
            },
            {
                "min": d + td(days=1),
                "max": None,
                "null_count": None,
                "len": None,
                "can_skip": True,
            },
        ],
    )


@given(
    s=series(
        name="x",
        min_size=1,
    ),
)
@settings(
    report_multiple_bugs=False,
    phases=(Phase.explicit, Phase.reuse, Phase.generate, Phase.target, Phase.explain),
)
def test_skip_batch_predicate_parametric(s: pl.Series) -> None:
    name = "x"
    dtype = s.dtype

    value_a = s.slice(0, 1)

    lit_a = pl.lit(value_a[0], dtype)

    exprs = [
        pl.col.x == lit_a,
        pl.col.x != lit_a,
        pl.col.x.eq_missing(lit_a),
        pl.col.x.ne_missing(lit_a),
        pl.col.x.is_null(),
        pl.col.x.is_not_null(),
    ]

    try:
        _ = s > value_a
        exprs += [
            pl.col.x > lit_a,
            pl.col.x >= lit_a,
            pl.col.x < lit_a,
            pl.col.x <= lit_a,
            pl.col.x.is_in(pl.Series([None, value_a[0]], dtype=dtype)),
        ]

        if s.len() > 1:
            value_b = s.slice(1, 1)
            lit_b = pl.lit(value_b[0], dtype)

            exprs += [
                pl.col.x.is_between(lit_a, lit_b),
                pl.col.x.is_in(pl.Series([value_a[0], value_b[0]], dtype=dtype)),
            ]
    except Exception as _:
        pass

    for expr in exprs:
        sbp = expr._skip_batch_predicate({name: dtype})

        if sbp is None:
            continue

        mins: list[PythonLiteral | None] = [None]
        with contextlib.suppress(Exception):
            mins = [s.min()]

        maxs: list[PythonLiteral | None] = [None]
        with contextlib.suppress(Exception):
            maxs = [s.max()]

        null_counts = [s.null_count()]
        lengths = [s.len()]

        df = pl.DataFrame(
            [
                pl.Series(f"{name}_min", mins, dtype),
                pl.Series(f"{name}_max", maxs, dtype),
                pl.Series(f"{name}_nc", null_counts, get_index_type()),
                pl.Series("len", lengths, get_index_type()),
            ]
        )

        can_skip = df.select(can_skip=sbp).fill_null(False).to_series()[0]
        if can_skip:
            try:
                assert s.to_frame().filter(expr).height == 0
            except Exception as _:
                print(expr)
                print(sbp)
                print(df)
                print(s.to_frame().filter(expr))

                raise



from __future__ import annotations

import io
from pathlib import Path
from uuid import uuid4

import numpy as np
import pytest

import polars as pl
from polars.exceptions import ComputeError
from polars.testing import assert_series_equal


def test_series_init_instantiated_object() -> None:
    s = pl.Series([object(), object()], dtype=pl.Object())
    assert isinstance(s, pl.Series)
    assert isinstance(s.dtype, pl.Object)


def test_object_empty_filter_5911() -> None:
    df = pl.DataFrame(
        data=[
            (1, "dog", {}),
        ],
        schema=[
            ("pet_id", pl.Int64),
            ("pet_type", pl.Categorical),
            ("pet_obj", pl.Object),
        ],
        orient="row",
    )

    empty_df = df.filter(pl.col("pet_type") == "cat")
    out = empty_df.select(["pet_obj"])
    assert out.dtypes == [pl.Object]
    assert out.shape == (0, 1)


def test_object_in_struct() -> None:
    np_a = np.array([1, 2, 3])
    np_b = np.array([4, 5, 6])
    df = pl.DataFrame({"A": [1, 2], "B": pl.Series([np_a, np_b], dtype=pl.Object)})

    with pytest.raises(pl.exceptions.InvalidOperationError):
        df.select([pl.struct(["B"])])


def test_nullable_object_13538() -> None:
    df = pl.DataFrame(
        data=[
            ({"a": 1},),
            ({"b": 3},),
            (None,),
        ],
        schema=[
            ("blob", pl.Object),
        ],
        orient="row",
    )

    df = df.select(
        is_null=pl.col("blob").is_null(), is_not_null=pl.col("blob").is_not_null()
    )
    assert df.to_dict(as_series=False) == {
        "is_null": [False, False, True],
        "is_not_null": [True, True, False],
    }

    df = pl.DataFrame({"col": pl.Series([0, 1, 2, None], dtype=pl.Object)})
    df = df.select(
        is_null=pl.col("col").is_null(), is_not_null=pl.col("col").is_not_null()
    )
    assert df.to_dict(as_series=False) == {
        "is_null": [False, False, False, True],
        "is_not_null": [True, True, True, False],
    }


@pytest.mark.may_fail_auto_streaming  # dtype is not set
def test_nullable_object_17936() -> None:
    class Custom:
        value: int

        def __init__(self, value: int) -> None:
            self.value = value

    def mapper(value: int) -> Custom | None:
        if value == 2:
            return None
        return Custom(value)

    df = pl.DataFrame({"a": [1, 2, 3]})

    assert df.select(
        pl.col("a").map_elements(mapper, return_dtype=pl.Object).alias("with_dtype"),
        pl.col("a").map_elements(mapper).alias("without_dtype"),
    ).null_count().row(0) == (1, 1)


def test_empty_sort() -> None:
    df = pl.DataFrame(
        data=[
            ({"name": "bar", "sort_key": 2},),
            ({"name": "foo", "sort_key": 1},),
        ],
        schema=[
            ("blob", pl.Object),
        ],
        orient="row",
    )
    df_filtered = df.filter(
        pl.col("blob").map_elements(
            lambda blob: blob["name"] == "baz", return_dtype=pl.Boolean
        )
    )
    df_filtered.sort(
        pl.col("blob").map_elements(
            lambda blob: blob["sort_key"], return_dtype=pl.Int64
        )
    )


def test_object_to_dicts() -> None:
    df = pl.DataFrame({"d": [{"a": 1, "b": 2, "c": 3}]}, schema={"d": pl.Object})
    assert df.to_dicts() == [{"d": {"a": 1, "b": 2, "c": 3}}]


def test_object_concat() -> None:
    df1 = pl.DataFrame(
        {"a": [1, 2, 3]},
        schema={"a": pl.Object},
    )

    df2 = pl.DataFrame(
        {"a": [1, 4, 3]},
        schema={"a": pl.Object},
    )

    catted = pl.concat([df1, df2])
    assert catted.shape == (6, 1)
    assert catted.dtypes == [pl.Object]
    assert catted.to_dict(as_series=False) == {"a": [1, 2, 3, 1, 4, 3]}


def test_object_concat_diagonal_14651() -> None:
    df1 = pl.DataFrame({"a": ["abc"]}, schema={"a": pl.Object})
    df2 = pl.DataFrame({"b": ["def"]}, schema={"b": pl.Object})
    result = pl.concat([df1, df2], how="diagonal")
    assert result.schema == pl.Schema({"a": pl.Object, "b": pl.Object})
    assert result["a"].to_list() == ["abc", None]
    assert result["b"].to_list() == [None, "def"]


def test_object_row_construction() -> None:
    data = [
        [uuid4()],
        [uuid4()],
        [uuid4()],
    ]
    df = pl.DataFrame(
        data,
        orient="row",
    )
    assert df.dtypes == [pl.Object]
    assert df["column_0"].to_list() == [value[0] for value in data]


def test_object_apply_to_struct() -> None:
    s = pl.Series([0, 1, 2], dtype=pl.Object)
    out = s.map_elements(lambda x: {"a": str(x), "b": x})
    assert out.dtype == pl.Struct([pl.Field("a", pl.String), pl.Field("b", pl.Int64)])


def test_null_obj_str_13512() -> None:
    # https://github.com/pola-rs/polars/issues/13512

    df1 = pl.DataFrame(
        {
            "key": [1],
        }
    )
    df2 = pl.DataFrame({"key": [2], "a": pl.Series([1], dtype=pl.Object)})

    out = df1.join(df2, on="key", how="left")
    s = str(out)
    assert s == (
        "shape: (1, 2)\n"
        "┌─────┬────────┐\n"
        "│ key ┆ a      │\n"
        "│ --- ┆ ---    │\n"
        "│ i64 ┆ object │\n"
        "╞═════╪════════╡\n"
        "│ 1   ┆ null   │\n"
        "└─────┴────────┘"
    )


def test_format_object_series_14267() -> None:
    s = pl.Series([Path(), Path("abc")])
    expected = "shape: (2,)\nSeries: '' [o][object]\n[\n\t.\n\tabc\n]"
    assert str(s) == expected


def test_object_raise_writers() -> None:
    df = pl.DataFrame({"a": object()})

    buf = io.BytesIO()

    with pytest.raises(ComputeError):
        df.write_parquet(buf)
    with pytest.raises(ComputeError):
        df.write_ipc(buf)
    with pytest.raises(ComputeError):
        df.write_json(buf)
    with pytest.raises(ComputeError):
        df.write_csv(buf)
    with pytest.raises(ComputeError):
        df.write_avro(buf)


def test_raise_list_object() -> None:
    # We don't want to support this. Unsafe enough as it is already.
    with pytest.raises(ValueError):
        pl.Series([[object()]], dtype=pl.List(pl.Object()))


def test_object_null_slice() -> None:
    s = pl.Series("x", [1, None, 42], dtype=pl.Object)
    assert_series_equal(s.is_null(), pl.Series("x", [False, True, False]))
    assert_series_equal(s.slice(0, 2).is_null(), pl.Series("x", [False, True]))
    assert_series_equal(s.slice(1, 1).is_null(), pl.Series("x", [True]))
    assert_series_equal(s.slice(2, 1).is_null(), pl.Series("x", [False]))


def test_object_sort_scalar_19925() -> None:
    a = object()
    assert pl.DataFrame({"a": [0], "obj": [a]}).sort("a")["obj"].item() == a


def test_object_estimated_size() -> None:
    df = pl.DataFrame(
        [
            ["3", "random python object, not a string"],
        ],
        schema={"name": pl.String, "ob": pl.Object},
        orient="row",
    )

    # is a huge underestimation
    assert df.estimated_size() == 9


def test_object_polars_dtypes_20572() -> None:
    df = pl.DataFrame(
        {
            "a": pl.Date(),
            "b": pl.Decimal(5, 1),
            "c": pl.Int64(),
            "d": pl.Object(),
            "e": pl.String(),
        }
    )
    assert all(dt.is_object() for dt in df.schema.dtypes())


def test_err_nested_object() -> None:
    a = object()
    df = pl.DataFrame({"obj": [a]})

    def mapper(x: object) -> list[object]:
        return [x]

    with pytest.raises(ValueError):
        df.select(pl.col("obj").map_elements(mapper))

    with pytest.raises(ValueError):
        df.select(pl.col("obj").map_elements(mapper, return_dtype=pl.List(pl.Object)))



from datetime import datetime
from typing import Any

import hypothesis.strategies as st
import pytest
from hypothesis import given, settings
from hypothesis.errors import InvalidArgument

import polars as pl
from polars.testing.parametric import (
    column,
    dataframes,
    dtypes,
    lists,
    series,
)

TEMPORAL_DTYPES = {pl.Date, pl.Time, pl.Datetime, pl.Duration}


@given(s=series())
@settings(max_examples=5)
def test_series_defaults(s: pl.Series) -> None:
    assert isinstance(s, pl.Series)
    assert s.name == ""


@given(s=series(name="hello"))
@settings(max_examples=5)
def test_series_name(s: pl.Series) -> None:
    assert s.name == "hello"


@given(st.data())
def test_series_dtype(data: st.DataObject) -> None:
    dtype = data.draw(dtypes())
    s = data.draw(series(dtype=dtype))
    assert s.dtype == dtype


@given(s=series(dtype=pl.Enum, allow_null=False))
@settings(max_examples=5)
def test_series_dtype_enum(s: pl.Series) -> None:
    assert isinstance(s.dtype, pl.Enum)
    assert all(v in s.dtype.categories for v in s)


@given(s=series(dtype=pl.Boolean, min_size=5, max_size=5))
@settings(max_examples=5)
def test_series_size(s: pl.Series) -> None:
    assert s.len() == 5


@given(s=series(min_size=3, max_size=8))
@settings(max_examples=5)
def test_series_size_range(s: pl.Series) -> None:
    assert 3 <= s.len() <= 8


@given(s=series(allow_null=False))
def test_series_allow_null_false(s: pl.Series) -> None:
    assert not s.has_nulls()
    assert s.dtype != pl.Null


@given(s=series(allowed_dtypes=[pl.Null], allow_null=False))
def test_series_allow_null_allowed_dtypes(s: pl.Series) -> None:
    assert s.dtype == pl.Null


@given(s=series(allowed_dtypes=[pl.List(pl.Int8)], allow_null=False))
def test_series_allow_null_nested(s: pl.Series) -> None:
    for v in s:
        assert not v.has_nulls()


@given(df=dataframes())
@settings(max_examples=5)
def test_dataframes_defaults(df: pl.DataFrame) -> None:
    assert isinstance(df, pl.DataFrame)
    assert df.columns == [f"col{i}" for i in range(df.width)]


@given(lf=dataframes(lazy=True))
@settings(max_examples=5)
def test_dataframes_lazy(lf: pl.LazyFrame) -> None:
    assert isinstance(lf, pl.LazyFrame)


@given(df=dataframes(cols=3, min_size=5, max_size=5))
@settings(max_examples=5)
def test_dataframes_size(df: pl.DataFrame) -> None:
    assert df.height == 5
    assert df.width == 3


@given(df=dataframes(min_cols=2, max_cols=5, min_size=3, max_size=8))
@settings(max_examples=5)
def test_dataframes_size_range(df: pl.DataFrame) -> None:
    assert 3 <= df.height <= 8
    assert 2 <= df.width <= 5


@given(df=dataframes(cols=1, allow_null=True))
@settings(max_examples=5)
def test_dataframes_allow_null_global(df: pl.DataFrame) -> None:
    null_count = sum(df.null_count().row(0))
    assert 0 <= null_count <= df.height * df.width


@given(df=dataframes(cols=2, allow_null={"col0": True}))
@settings(max_examples=5)
def test_dataframes_allow_null_column(df: pl.DataFrame) -> None:
    null_count = sum(df.null_count().row(0))
    assert 0 <= null_count <= df.height * df.width


@given(
    df=dataframes(
        cols=1,
        allow_null=False,
        include_cols=[column(name="colx", allow_null=True)],
    )
)
def test_dataframes_allow_null_override(df: pl.DataFrame) -> None:
    assert not df.get_column("col0").has_nulls()
    assert 0 <= df.get_column("colx").null_count() <= df.height


@given(
    lf=dataframes(
        # generate lazyframes with at least one row
        lazy=True,
        min_size=1,
        allow_null=False,
        # test mix & match of bulk-assigned cols with custom cols
        cols=[column(n, dtype=pl.UInt8, unique=True) for n in ["a", "b"]],
        include_cols=[
            column("c", dtype=pl.Boolean),
            column("d", strategy=st.sampled_from(["x", "y", "z"])),
        ],
    )
)
def test_dataframes_columns(lf: pl.LazyFrame) -> None:
    assert lf.collect_schema() == {
        "a": pl.UInt8,
        "b": pl.UInt8,
        "c": pl.Boolean,
        "d": pl.String,
    }
    assert lf.collect_schema().names() == ["a", "b", "c", "d"]
    df = lf.collect()

    # confirm uint cols bounds
    uint8_max = (2**8) - 1
    assert df["a"].min() >= 0  # type: ignore[operator]
    assert df["b"].min() >= 0  # type: ignore[operator]
    assert df["a"].max() <= uint8_max  # type: ignore[operator]
    assert df["b"].max() <= uint8_max  # type: ignore[operator]

    # confirm uint cols uniqueness
    assert df["a"].is_unique().all()
    assert df["b"].is_unique().all()

    # boolean col
    assert all(isinstance(v, bool) for v in df["c"].to_list())

    # string col, entries selected from custom values
    xyz = {"x", "y", "z"}
    assert all(v in xyz for v in df["d"].to_list())


@pytest.mark.hypothesis
def test_column_invalid_probability() -> None:
    with pytest.deprecated_call(), pytest.raises(InvalidArgument):
        column("col", null_probability=2.0)


@pytest.mark.hypothesis
def test_column_null_probability_deprecated() -> None:
    with pytest.deprecated_call():
        col = column("col", allow_null=False, null_probability=0.5)
    assert col.null_probability == 0.5
    assert col.allow_null is True  # null_probability takes precedence


@given(st.data())
def test_allow_infinities_deprecated(data: st.DataObject) -> None:
    with pytest.deprecated_call():
        strategy = series(dtype=pl.Float64, allow_infinities=False)
        s = data.draw(strategy)

    assert all(v not in (float("inf"), float("-inf")) for v in s)


@given(
    df=dataframes(
        cols=[
            column("colx", dtype=pl.Array(pl.UInt8, shape=3)),
            column("coly", dtype=pl.List(pl.Datetime("ms"))),
            column(
                name="colz",
                dtype=pl.List(pl.List(pl.String)),
                strategy=lists(
                    inner_dtype=pl.List(pl.String),
                    select_from=["aa", "bb", "cc"],
                    min_size=1,
                ),
            ),
        ],
        allow_null=False,
    ),
)
def test_dataframes_nested_strategies(df: pl.DataFrame) -> None:
    assert df.schema == {
        "colx": pl.Array(pl.UInt8, shape=3),
        "coly": pl.List(pl.Datetime("ms")),
        "colz": pl.List(pl.List(pl.String)),
    }
    uint8_max = (2**8) - 1

    for colx, coly, colz in df.iter_rows():
        assert len(colx) == 3
        assert all(i <= uint8_max for i in colx)
        assert all(isinstance(d, datetime) for d in coly)
        for inner_list in colz:
            assert all(s in ("aa", "bb", "cc") for s in inner_list)


@given(
    df=dataframes(allowed_dtypes=TEMPORAL_DTYPES, max_size=1, max_cols=5),
    lf=dataframes(excluded_dtypes=TEMPORAL_DTYPES, max_size=1, max_cols=5, lazy=True),
    s1=series(allowed_dtypes=TEMPORAL_DTYPES, max_size=1),
    s2=series(excluded_dtypes=TEMPORAL_DTYPES, max_size=1),
)
@settings(max_examples=50)
def test_strategy_dtypes(
    df: pl.DataFrame,
    lf: pl.LazyFrame,
    s1: pl.Series,
    s2: pl.Series,
) -> None:
    # dataframe, lazyframe
    assert all(tp.is_temporal() for tp in df.dtypes)
    assert all(not tp.is_temporal() for tp in lf.collect_schema().dtypes())

    # series
    assert s1.dtype.is_temporal()
    assert not s2.dtype.is_temporal()


@given(s=series(allow_chunks=False))
@settings(max_examples=10)
@pytest.mark.may_fail_cloud
def test_series_allow_chunks(s: pl.Series) -> None:
    assert s.n_chunks() == 1


@given(df=dataframes(allow_chunks=False))
@settings(max_examples=10)
@pytest.mark.may_fail_cloud
def test_dataframes_allow_chunks(df: pl.DataFrame) -> None:
    assert df.n_chunks("first") == 1
    assert df.n_chunks("all") == [1] * df.width


@given(
    df=dataframes(
        allowed_dtypes=[pl.Float32, pl.Float64],
        max_cols=4,
        allow_null=False,
        allow_infinity=False,
    ),
    s=series(dtype=pl.Float64, allow_null=False, allow_infinity=False),
)
def test_infinities(
    df: pl.DataFrame,
    s: pl.Series,
) -> None:
    from math import isfinite, isnan

    def finite_float(value: Any) -> bool:
        return isfinite(value) or isnan(value)

    assert all(finite_float(val) for val in s.to_list())
    for col in df.columns:
        assert all(finite_float(val) for val in df[col].to_list())


@given(
    df=dataframes(
        cols=10,
        max_size=1,
        allowed_dtypes=[pl.Int8, pl.UInt16, pl.List(pl.Int32)],
    )
)
@settings(max_examples=3)
def test_dataframes_allowed_dtypes_integer_cols(df: pl.DataFrame) -> None:
    # ensure dtype constraint works in conjunction with 'n' cols
    assert all(
        tp in (pl.Int8, pl.UInt16, pl.List(pl.Int32)) for tp in df.schema.values()
    )


@given(st.data())
@settings(max_examples=1)
def test_series_chunked_deprecated(data: st.DataObject) -> None:
    with pytest.deprecated_call():
        data.draw(series(chunked=True))
    with pytest.deprecated_call():
        data.draw(dataframes(chunked=True))



from __future__ import annotations

from datetime import date, datetime
from typing import TYPE_CHECKING

import pytest

import polars as pl
from polars.exceptions import ComputeError
from polars.testing import assert_frame_equal, assert_series_equal

if TYPE_CHECKING:
    from polars._typing import TimeUnit


def test_replace_expr_datetime() -> None:
    df = pl.DataFrame(
        {
            "dates": [
                datetime(2088, 8, 8, 8, 8, 8, 8),
                datetime(2088, 8, 8, 8, 8, 8, 8),
                datetime(2088, 8, 8, 8, 8, 8, 8),
                datetime(2088, 8, 8, 8, 8, 8, 8),
                datetime(2088, 8, 8, 8, 8, 8, 8),
                datetime(2088, 8, 8, 8, 8, 8, 8),
                datetime(2088, 8, 8, 8, 8, 8, 8),
                None,
            ],
            "year": [None, 2, 3, 4, 5, 6, 7, 8],
            "month": [1, None, 3, 4, 5, 6, 7, 8],
            "day": [1, 2, None, 4, 5, 6, 7, 8],
            "hour": [1, 2, 3, None, 5, 6, 7, 8],
            "minute": [1, 2, 3, 4, None, 6, 7, 8],
            "second": [1, 2, 3, 4, 5, None, 7, 8],
            "microsecond": [1, 2, 3, 4, 5, 6, None, 8],
        }
    )

    result = df.select(
        pl.col("dates").dt.replace(
            year="year",
            month="month",
            day="day",
            hour="hour",
            minute="minute",
            second="second",
            microsecond="microsecond",
        )
    )

    expected = pl.DataFrame(
        {
            "dates": [
                datetime(2088, 1, 1, 1, 1, 1, 1),
                datetime(2, 8, 2, 2, 2, 2, 2),
                datetime(3, 3, 8, 3, 3, 3, 3),
                datetime(4, 4, 4, 8, 4, 4, 4),
                datetime(5, 5, 5, 5, 8, 5, 5),
                datetime(6, 6, 6, 6, 6, 8, 6),
                datetime(7, 7, 7, 7, 7, 7, 8),
                None,
            ]
        }
    )

    assert_frame_equal(result, expected)


def test_replace_expr_date() -> None:
    df = pl.DataFrame(
        {
            "dates": [date(2088, 8, 8), date(2088, 8, 8), date(2088, 8, 8), None],
            "year": [None, 2, 3, 4],
            "month": [1, None, 3, 4],
            "day": [1, 2, None, 4],
        }
    )

    result = df.select(
        pl.col("dates").dt.replace(year="year", month="month", day="day")
    )

    expected = pl.DataFrame(
        {"dates": [date(2088, 1, 1), date(2, 8, 2), date(3, 3, 8), None]}
    )

    assert_frame_equal(result, expected)


def test_replace_int_datetime() -> None:
    df = pl.DataFrame(
        {
            "a": [
                datetime(1, 1, 1, 1, 1, 1, 1),
                datetime(2, 2, 2, 2, 2, 2, 2),
                datetime(3, 3, 3, 3, 3, 3, 3),
                None,
            ]
        }
    )
    result = df.select(
        pl.col("a").dt.replace().alias("no_change"),
        pl.col("a").dt.replace(year=9).alias("year"),
        pl.col("a").dt.replace(month=9).alias("month"),
        pl.col("a").dt.replace(day=9).alias("day"),
        pl.col("a").dt.replace(hour=9).alias("hour"),
        pl.col("a").dt.replace(minute=9).alias("minute"),
        pl.col("a").dt.replace(second=9).alias("second"),
        pl.col("a").dt.replace(microsecond=9).alias("microsecond"),
    )
    expected = pl.DataFrame(
        {
            "no_change": [
                datetime(1, 1, 1, 1, 1, 1, 1),
                datetime(2, 2, 2, 2, 2, 2, 2),
                datetime(3, 3, 3, 3, 3, 3, 3),
                None,
            ],
            "year": [
                datetime(9, 1, 1, 1, 1, 1, 1),
                datetime(9, 2, 2, 2, 2, 2, 2),
                datetime(9, 3, 3, 3, 3, 3, 3),
                None,
            ],
            "month": [
                datetime(1, 9, 1, 1, 1, 1, 1),
                datetime(2, 9, 2, 2, 2, 2, 2),
                datetime(3, 9, 3, 3, 3, 3, 3),
                None,
            ],
            "day": [
                datetime(1, 1, 9, 1, 1, 1, 1),
                datetime(2, 2, 9, 2, 2, 2, 2),
                datetime(3, 3, 9, 3, 3, 3, 3),
                None,
            ],
            "hour": [
                datetime(1, 1, 1, 9, 1, 1, 1),
                datetime(2, 2, 2, 9, 2, 2, 2),
                datetime(3, 3, 3, 9, 3, 3, 3),
                None,
            ],
            "minute": [
                datetime(1, 1, 1, 1, 9, 1, 1),
                datetime(2, 2, 2, 2, 9, 2, 2),
                datetime(3, 3, 3, 3, 9, 3, 3),
                None,
            ],
            "second": [
                datetime(1, 1, 1, 1, 1, 9, 1),
                datetime(2, 2, 2, 2, 2, 9, 2),
                datetime(3, 3, 3, 3, 3, 9, 3),
                None,
            ],
            "microsecond": [
                datetime(1, 1, 1, 1, 1, 1, 9),
                datetime(2, 2, 2, 2, 2, 2, 9),
                datetime(3, 3, 3, 3, 3, 3, 9),
                None,
            ],
        }
    )
    assert_frame_equal(result, expected)


def test_replace_int_date() -> None:
    df = pl.DataFrame(
        {
            "a": [
                date(1, 1, 1),
                date(2, 2, 2),
                date(3, 3, 3),
                None,
            ]
        }
    )
    result = df.select(
        pl.col("a").dt.replace().alias("no_change"),
        pl.col("a").dt.replace(year=9).alias("year"),
        pl.col("a").dt.replace(month=9).alias("month"),
        pl.col("a").dt.replace(day=9).alias("day"),
    )
    expected = pl.DataFrame(
        {
            "no_change": [
                date(1, 1, 1),
                date(2, 2, 2),
                date(3, 3, 3),
                None,
            ],
            "year": [
                date(9, 1, 1),
                date(9, 2, 2),
                date(9, 3, 3),
                None,
            ],
            "month": [
                date(1, 9, 1),
                date(2, 9, 2),
                date(3, 9, 3),
                None,
            ],
            "day": [
                date(1, 1, 9),
                date(2, 2, 9),
                date(3, 3, 9),
                None,
            ],
        }
    )
    assert_frame_equal(result, expected)


def test_replace_ambiguous() -> None:
    # Value to be replaced by an ambiguous hour.
    value = pl.select(
        pl.datetime(2020, 10, 25, 5, time_zone="Europe/London")
    ).to_series()

    input = [2020, 10, 25, 1]
    tz = "Europe/London"

    # earliest
    expected = pl.select(
        pl.datetime(*input, time_zone=tz, ambiguous="earliest")
    ).to_series()
    result = value.dt.replace(hour=1, ambiguous="earliest")
    assert_series_equal(result, expected)

    # latest
    expected = pl.select(
        pl.datetime(*input, time_zone=tz, ambiguous="latest")
    ).to_series()
    result = value.dt.replace(hour=1, ambiguous="latest")
    assert_series_equal(result, expected)

    # null
    expected = pl.select(
        pl.datetime(*input, time_zone=tz, ambiguous="null")
    ).to_series()
    result = value.dt.replace(hour=1, ambiguous="null")
    assert_series_equal(result, expected)

    # raise
    with pytest.raises(
        ComputeError,
        match=(
            "datetime '2020-10-25 01:00:00' is ambiguous in time zone 'Europe/London'. "
            "Please use `ambiguous` to tell how it should be localized."
        ),
    ):
        value.dt.replace(hour=1, ambiguous="raise")


def test_replace_datetime_preserve_ns() -> None:
    df = pl.DataFrame(
        {
            "a": pl.Series(["2020-01-01T00:00:00.123456789"] * 2).cast(
                pl.Datetime("ns")
            ),
            "year": [2021, None],
            "microsecond": [50, None],
        }
    )

    result = df.select(
        year=pl.col("a").dt.replace(year="year"),
        us=pl.col("a").dt.replace(microsecond="microsecond"),
    )

    expected = pl.DataFrame(
        {
            "year": pl.Series(
                [
                    "2021-01-01T00:00:00.123456789",
                    "2020-01-01T00:00:00.123456789",
                ]
            ).cast(pl.Datetime("ns")),
            "us": pl.Series(
                [
                    "2020-01-01T00:00:00.000050",
                    "2020-01-01T00:00:00.123456789",
                ]
            ).cast(pl.Datetime("ns")),
        }
    )

    assert_frame_equal(result, expected)


@pytest.mark.parametrize("tu", ["ms", "us", "ns"])
@pytest.mark.parametrize("tzinfo", [None, "Africa/Nairobi", "America/New_York"])
def test_replace_preserve_tu_and_tz(tu: TimeUnit, tzinfo: str) -> None:
    s = pl.Series(
        [datetime(2024, 1, 1), datetime(2024, 1, 2)],
        dtype=pl.Datetime(time_unit=tu, time_zone=tzinfo),
    )
    result = s.dt.replace(year=2000)
    assert result.dtype.time_unit == tu  # type: ignore[attr-defined]
    assert result.dtype.time_zone == tzinfo  # type: ignore[attr-defined]


def test_replace_date_invalid_components() -> None:
    df = pl.DataFrame({"a": [date(2025, 1, 1)]})

    with pytest.raises(
        ComputeError, match=r"Invalid date components \(2025, 13, 1\) supplied"
    ):
        df.select(pl.col("a").dt.replace(month=13))
    with pytest.raises(
        ComputeError, match=r"Invalid date components \(2025, 1, 32\) supplied"
    ):
        df.select(pl.col("a").dt.replace(day=32))


def test_replace_datetime_invalid_date_components() -> None:
    df = pl.DataFrame({"a": [datetime(2025, 1, 1)]})

    with pytest.raises(
        ComputeError, match=r"Invalid date components \(2025, 13, 1\) supplied"
    ):
        df.select(pl.col("a").dt.replace(month=13))
    with pytest.raises(
        ComputeError, match=r"Invalid date components \(2025, 1, 32\) supplied"
    ):
        df.select(pl.col("a").dt.replace(day=32))


def test_replace_datetime_invalid_time_components() -> None:
    df = pl.DataFrame({"a": [datetime(2025, 1, 1)]})

    # hour
    with pytest.raises(
        ComputeError, match=r"Invalid time components \(25, 0, 0, 0\) supplied"
    ):
        df.select(pl.col("a").dt.replace(hour=25))

    # minute
    with pytest.raises(
        ComputeError, match=r"Invalid time components \(0, 61, 0, 0\) supplied"
    ):
        df.select(pl.col("a").dt.replace(minute=61))

    # second
    with pytest.raises(
        ComputeError, match=r"Invalid time components \(0, 0, 61, 0\) supplied"
    ):
        df.select(pl.col("a").dt.replace(second=61))

    # microsecond
    with pytest.raises(
        ComputeError,
        match=r"Invalid time components \(0, 0, 0, 2000000000\) supplied",
    ):
        df.select(pl.col("a").dt.replace(microsecond=2_000_000))


def test_replace_unequal_length_22018() -> None:
    with pytest.raises(pl.exceptions.ShapeError):
        pl.Series([datetime(2088, 8, 8, 8, 8, 8, 8)] * 2).dt.replace(
            year=pl.Series([2000, 2001, 2002])
        )


def test_replace_broadcast_self() -> None:
    df = pl.DataFrame(
        {
            "year": [None, 2, 3, 4, 5, 6, 7, 8],
            "month": [1, None, 3, 4, 5, 6, 7, 8],
            "day": [1, 2, None, 4, 5, 6, 7, 8],
            "hour": [1, 2, 3, None, 5, 6, 7, 8],
            "minute": [1, 2, 3, 4, None, 6, 7, 8],
            "second": [1, 2, 3, 4, 5, None, 7, 8],
            "microsecond": [1, 2, 3, 4, 5, 6, None, 8],
        }
    )

    result = df.select(
        pl.lit(pl.Series("dates", [datetime(2088, 8, 8, 8, 8, 8, 8)])).dt.replace(
            year="year",
            month="month",
            day="day",
            hour="hour",
            minute="minute",
            second="second",
            microsecond="microsecond",
        )
    )

    expected = pl.DataFrame(
        {
            "dates": [
                datetime(2088, 1, 1, 1, 1, 1, 1),
                datetime(2, 8, 2, 2, 2, 2, 2),
                datetime(3, 3, 8, 3, 3, 3, 3),
                datetime(4, 4, 4, 8, 4, 4, 4),
                datetime(5, 5, 5, 5, 8, 5, 5),
                datetime(6, 6, 6, 6, 6, 8, 6),
                datetime(7, 7, 7, 7, 7, 7, 8),
                datetime(8, 8, 8, 8, 8, 8, 8),
            ]
        }
    )

    assert_frame_equal(result, expected)



from __future__ import annotations

import datetime

import pytest

import polars as pl


def test_series_item() -> None:
    s = pl.Series("a", [1])
    assert s.item() == 1


def test_series_item_empty() -> None:
    s = pl.Series("a", [])
    with pytest.raises(ValueError):
        s.item()


def test_series_item_incorrect_shape() -> None:
    s = pl.Series("a", [1, 2])
    with pytest.raises(ValueError):
        s.item()


@pytest.fixture(scope="module")
def s() -> pl.Series:
    return pl.Series("a", [1, 2])


@pytest.mark.parametrize(("index", "expected"), [(0, 1), (1, 2), (-1, 2), (-2, 1)])
def test_series_item_with_index(index: int, expected: int, s: pl.Series) -> None:
    assert s.item(index) == expected


@pytest.mark.parametrize("index", [-10, 10])
def test_df_item_out_of_bounds(index: int, s: pl.Series) -> None:
    with pytest.raises(IndexError, match="out of bounds"):
        s.item(index)


def test_series_item_out_of_range_date() -> None:
    s = pl.Series([datetime.date(9999, 12, 31)]).dt.offset_by("1d")
    with pytest.raises(ValueError, match="out of range"):
        s.item()



from __future__ import annotations

import subprocess
import sys
from typing import TYPE_CHECKING

import pytest

import polars as pl
from polars.testing import assert_frame_equal

if TYPE_CHECKING:
    from pathlib import Path


@pytest.fixture(scope="session")
def data_files_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    """Creates: 5 parquet files, 5 rows each."""
    tmp_path = tmp_path_factory.mktemp("data-files")
    _create_data_files(tmp_path)

    return tmp_path


@pytest.fixture
def write_position_deletes(tmp_path: Path) -> WritePositionDeletes:
    return WritePositionDeletes(tmp_path=tmp_path)


class WritePositionDeletes:  # noqa: D101
    def __init__(self, *, tmp_path: Path) -> None:
        self.tmp_path = tmp_path
        self.i = 0

    def __call__(self, positions: pl.Series) -> str:
        path = self.tmp_path / f"{self.i}"

        (
            positions.alias("pos")
            .to_frame()
            .select(pl.lit("").alias("file_path"), "pos")
            .write_parquet(path)
        )

        self.i += 1

        return str(path)


# 5 files x 5 rows each. Contains `physical_index` [0, 1, .., 24].
def _create_data_files(tmp_path: Path) -> None:
    tmp_path.mkdir(exist_ok=True, parents=True)
    df = pl.select(physical_index=pl.int_range(25, dtype=pl.UInt32))

    parts = []

    for i in [0, 5, 10, 15, 20]:
        O = "0" if i < 10 else ""  # noqa: E741
        path = tmp_path / f"offset={O}{i}/data.parquet"
        path.parent.mkdir(exist_ok=True, parents=True)

        part_df = df.slice(i, 5)
        part_df.write_parquet(path)

        parts.append(part_df.with_columns(offset=pl.lit(i, dtype=pl.Int64)))

    assert_frame_equal(pl.scan_parquet(tmp_path).collect(), pl.concat(parts))


@pytest.mark.parametrize("row_index_offset", [0, 27, 38, 73])
def test_scan_row_deletions(
    data_files_path: Path,
    write_position_deletes: WritePositionDeletes,
    row_index_offset: int,
) -> None:
    deletion_files = (
        "iceberg-position-delete",
        {
            0: [
                write_position_deletes(pl.Series([1, 2])),
            ],
            1: [
                write_position_deletes(pl.Series([0, 1, 2])),
            ],
            4: [
                write_position_deletes(pl.Series([2, 3])),
            ],
        },
    )

    def apply_row_index_offset(values: list[int]) -> list[int]:
        return [x + row_index_offset for x in values]

    q = pl.scan_parquet(
        data_files_path,
        _deletion_files=deletion_files,  # type: ignore[arg-type]
        hive_partitioning=False,
    ).with_row_index(offset=row_index_offset)

    assert q.select(pl.len()).collect().item() == 18

    assert_frame_equal(
        q.collect(),
        pl.DataFrame(
            {
                "index": apply_row_index_offset([
                    0, 1, 2,
                    3, 4,
                    5, 6, 7, 8, 9,
                    10, 11, 12, 13, 14,
                    15, 16, 17
                ]),
                "physical_index": [
                    0, 3, 4,
                    8, 9,
                    10, 11, 12, 13, 14,
                    15, 16, 17, 18, 19,
                    20, 21, 24
                ],
            },
            schema={"index": pl.get_index_type(), "physical_index": pl.UInt32},
        )
    )  # fmt: skip

    # head()

    assert_frame_equal(
        q.head(3).collect(),
        pl.DataFrame(
            {
                "index": apply_row_index_offset([0, 1, 2]),
                "physical_index": [0, 3, 4],
            },
            schema={"index": pl.get_index_type(), "physical_index": pl.UInt32},
        ),
    )

    assert_frame_equal(
        q.head(10).collect(),
        pl.DataFrame(
            {
                "index": apply_row_index_offset([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
                "physical_index": [0, 3, 4, 8, 9, 10, 11, 12, 13, 14],
            },
            schema={"index": pl.get_index_type(), "physical_index": pl.UInt32},
        ),
    )

    # tail()

    assert_frame_equal(
        q.tail(3).collect(),
        pl.DataFrame(
            {
                "index": apply_row_index_offset([15, 16, 17]),
                "physical_index": [20, 21, 24],
            },
            schema={"index": pl.get_index_type(), "physical_index": pl.UInt32},
        ),
    )

    assert_frame_equal(
        q.tail(10).collect(),
        pl.DataFrame(
            {
                "index": apply_row_index_offset(
                    [8, 9, 10, 11, 12, 13, 14, 15, 16, 17]
                ),
                "physical_index": [
                    13, 14, 15, 16, 17, 18, 19, 20, 21,
                    24
                ],
            },
            schema={"index": pl.get_index_type(), "physical_index": pl.UInt32},
        ),
    )  # fmt: skip

    # slice(positive_offset)

    assert_frame_equal(
        q.slice(2, 10).collect(),
        pl.DataFrame(
            {
                "index": apply_row_index_offset([2, 3, 4, 5, 6, 7, 8, 9, 10, 11]),
                "physical_index": [4, 8, 9, 10, 11, 12, 13, 14, 15, 16],
            },
            schema={"index": pl.get_index_type(), "physical_index": pl.UInt32},
        ),
    )

    assert_frame_equal(
        q.slice(5, 10).collect(),
        pl.DataFrame(
            {
                "index": apply_row_index_offset([5, 6, 7, 8, 9, 10, 11, 12, 13, 14]),
                "physical_index": [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
            },
            schema={"index": pl.get_index_type(), "physical_index": pl.UInt32},
        ),
    )

    assert_frame_equal(
        q.slice(10, 10).collect(),
        pl.DataFrame(
            {
                "index": apply_row_index_offset([10, 11, 12, 13, 14, 15, 16, 17]),
                "physical_index": [15, 16, 17, 18, 19, 20, 21, 24],
            },
            schema={"index": pl.get_index_type(), "physical_index": pl.UInt32},
        ),
    )

    # slice(negative_offset)
    assert_frame_equal(
        q.slice(-3, 2).collect(),
        pl.DataFrame(
            {
                "index": apply_row_index_offset([15, 16]),
                "physical_index": [20, 21],
            },
            schema={"index": pl.get_index_type(), "physical_index": pl.UInt32},
        ),
    )

    assert_frame_equal(
        q.slice(-23, 10).collect(),
        pl.DataFrame(
            {
                "index": apply_row_index_offset([0, 1, 2, 3, 4]),
                "physical_index": [0, 3, 4, 8, 9],
            },
            schema={"index": pl.get_index_type(), "physical_index": pl.UInt32},
        ),
    )

    # filter: skip_files
    q = pl.scan_parquet(
        data_files_path,
        _deletion_files=deletion_files,  # type: ignore[arg-type]
    ).with_row_index(offset=row_index_offset)

    assert_frame_equal(
        q.filter(pl.col("offset").is_in([10, 20])).collect(),
        pl.DataFrame(
            {
                "index": apply_row_index_offset([5, 6, 7, 8, 9, 15, 16, 17]),
                "physical_index": [10, 11, 12, 13, 14, 20, 21, 24],
                "offset": [10, 10, 10, 10, 10, 20, 20, 20],
            },
            schema={
                "index": pl.get_index_type(),
                "physical_index": pl.UInt32,
                "offset": pl.Int64,
            },
        ),
    )


@pytest.mark.slow
@pytest.mark.write_disk
@pytest.mark.parametrize("ideal_morsel_size", [999, 50, 33])
@pytest.mark.parametrize("force_empty_capabilities", [True, False])
def test_scan_row_deletion_single_large(
    tmp_path: Path,
    write_position_deletes: WritePositionDeletes,
    ideal_morsel_size: int,
    force_empty_capabilities: bool,
) -> None:
    path = tmp_path / "data.parquet"
    pl.DataFrame({"physical_index": range(100)}).write_parquet(path)

    positions = pl.Series([
        0,   1,  2,  3,  4,  5,  6,  7,  8, 22,
        23, 24, 25, 26, 27, 28, 29, 30, 31, 32,
        33, 45, 46, 47, 48, 49, 50, 51, 52, 53,
        54, 64, 65, 66, 67, 68, 69, 70, 71, 79,
        80, 81, 82, 83, 84, 90, 91, 92, 93, 97,
        98
    ])  # fmt: skip

    deletion_positions_path = write_position_deletes(positions)

    script_args: list[str] = [
        str(ideal_morsel_size),
        "1" if force_empty_capabilities else "0",
        str(path),
        deletion_positions_path,
    ]

    # Use a process to ensure ideal morsel size is set correctly.
    out = subprocess.check_output(
        [
            sys.executable,
            "-c",
            """\
import os
import sys

(
    _,
    ideal_morsel_size,
    force_empty_capabilities,
    data_file_path,
    deletion_positions_path,
) = sys.argv

os.environ["POLARS_VERBOSE"] = "0"
os.environ["POLARS_MAX_THREADS"] = "1"
os.environ["POLARS_IDEAL_MORSEL_SIZE"] = ideal_morsel_size
os.environ["POLARS_FORCE_EMPTY_READER_CAPABILITIES"] = force_empty_capabilities

import polars as pl
from polars.testing import assert_frame_equal

full_expected_physical = [
     9, 10, 11, 12, 13, 14, 15, 16, 17, 18,
    19, 20, 21, 34, 35, 36, 37, 38, 39, 40,
    41, 42, 43, 44, 55, 56, 57, 58, 59, 60,
    61, 62, 63, 72, 73, 74, 75, 76, 77, 78,
    85, 86, 87, 88, 89, 94, 95, 96, 99
]  # fmt: skip

deletion_files = (
    "iceberg-position-delete",
    {0: [deletion_positions_path]},
)

q = pl.scan_parquet(data_file_path, _deletion_files=deletion_files).with_row_index()

assert_frame_equal(
    q.collect(),
    pl.DataFrame({"physical_index": full_expected_physical}).with_row_index(),
)

assert_frame_equal(
    q.tail(999).collect(),
    pl.DataFrame({"physical_index": full_expected_physical}).with_row_index(),
)

# Note: The negative slice is important here. Otherwise row_index does not get
# lowered into the post-apply pipeline.
for negative_offset in range(1, 49):
    assert_frame_equal(
        q.tail(negative_offset).collect(),
        pl.DataFrame(
            {"physical_index": full_expected_physical[-negative_offset:]}
        ).with_row_index(offset=49 - negative_offset),
    )

assert_frame_equal(
    q.slice(20).collect(),
    pl.DataFrame({"physical_index": full_expected_physical[20:]}).with_row_index(
        offset=20
    ),
)

print("OK", end="")
    """,
            *script_args,
        ],
        stderr=subprocess.STDOUT,
    )

    assert out == b"OK"


@pytest.mark.write_disk
def test_scan_row_deletion_skips_file_with_all_rows_deleted(
    tmp_path: Path,
    write_position_deletes: WritePositionDeletes,
) -> None:
    # Create our own copy because we mutate one of the data files
    data_files_path = tmp_path / "data-files"
    _create_data_files(data_files_path)

    # Corrupt a parquet file
    def remove_data(path: Path) -> None:
        v = path.read_bytes()
        metadata_and_footer_len = 8 + int.from_bytes(v[-8:][:4], "little")
        path.write_bytes(b"\x00" * (len(v) - metadata_and_footer_len))
        path.write_bytes(v[-metadata_and_footer_len:])

    remove_data(data_files_path / "offset=05/data.parquet")

    q = pl.scan_parquet(
        data_files_path / "offset=05/data.parquet", hive_partitioning=False
    )

    # Baseline: The metadata is readable but the row groups are not

    assert q.collect_schema() == {"physical_index": pl.UInt32}
    assert q.select(pl.len()).collect().item() == 5

    with pytest.raises(pl.exceptions.ComputeError, match="Invalid thrift"):
        q.collect()

    q = pl.scan_parquet(data_files_path, hive_partitioning=False)

    with pytest.raises(pl.exceptions.ComputeError, match="Invalid thrift"):
        q.collect()

    q = pl.scan_parquet(
        data_files_path,
        _deletion_files=(
            "iceberg-position-delete",
            {
                1: [
                    write_position_deletes(pl.Series([0, 1, 2])),
                    write_position_deletes(pl.Series([3, 4])),
                ]
            },
        ),
        hive_partitioning=False,
    )

    expect = pl.DataFrame(
        {
            "index": [
                0, 1, 2, 3, 4,
                5, 6, 7, 8, 9,
                10, 11, 12, 13, 14,
                15, 16, 17, 18, 19,
            ],
            "physical_index": [
                0, 1, 2, 3, 4,
                10, 11, 12, 13, 14,
                15, 16, 17, 18, 19,
                20, 21, 22, 23, 24,
            ],
        },
        schema={"index": pl.get_index_type(), "physical_index": pl.UInt32},
    )  # fmt: skip

    assert_frame_equal(q.collect(), expect.drop("index"))
    assert_frame_equal(q.with_row_index().collect(), expect)

    expect = pl.DataFrame(
        {
            "index": [
                10, 11, 12, 13, 14,
                15, 16, 17, 18, 19,
            ],
            "physical_index": [
                15, 16, 17, 18, 19,
                20, 21, 22, 23, 24,
            ],
        },
        schema={"index": pl.get_index_type(), "physical_index": pl.UInt32},
    )  # fmt: skip

    assert_frame_equal(q.slice(10).collect(), expect.drop("index"))
    assert_frame_equal(q.with_row_index().slice(10).collect(), expect)



import itertools

import pytest

import polars as pl
from polars.testing import assert_frame_equal


def test_is_null_followed_by_all() -> None:
    lf = pl.LazyFrame({"group": [0, 0, 0, 1], "val": [6, 0, None, None]})

    expected_df = pl.DataFrame({"group": [0, 1], "val": [False, True]})
    result_lf = lf.group_by("group", maintain_order=True).agg(
        pl.col("val").is_null().all()
    )

    assert r'[[(col("val").len()) == (col("val").null_count())]]' in result_lf.explain()
    assert "is_null" not in result_lf
    assert_frame_equal(expected_df, result_lf.collect())

    # verify we don't optimize on chained expressions when last one is not col
    non_optimized_result_plan = (
        lf.group_by("group", maintain_order=True)
        .agg(pl.col("val").abs().is_null().all())
        .explain()
    )
    assert "null_count" not in non_optimized_result_plan
    assert "is_null" in non_optimized_result_plan

    # edge case of empty series
    lf = pl.LazyFrame({"val": []}, schema={"val": pl.Int32})

    expected_df = pl.DataFrame({"val": [True]})
    result_df = lf.select(pl.col("val").is_null().all()).collect()
    assert_frame_equal(expected_df, result_df)


def test_is_null_followed_by_any() -> None:
    lf = pl.LazyFrame({"group": [0, 0, 0, 1, 2], "val": [6, 0, None, None, 5]})

    expected_df = pl.DataFrame({"group": [0, 1, 2], "val": [True, True, False]})
    result_lf = lf.group_by("group", maintain_order=True).agg(
        pl.col("val").is_null().any()
    )
    assert_frame_equal(expected_df, result_lf.collect())

    # edge case of empty series
    lf = pl.LazyFrame({"val": []}, schema={"val": pl.Int32})

    expected_df = pl.DataFrame({"val": [False]})
    result_df = lf.select(pl.col("val").is_null().any()).collect()
    assert_frame_equal(expected_df, result_df)


def test_is_not_null_followed_by_all() -> None:
    lf = pl.LazyFrame({"group": [0, 0, 0, 1], "val": [6, 0, 5, None]})

    expected_df = pl.DataFrame({"group": [0, 1], "val": [True, False]})
    result_df = (
        lf.group_by("group", maintain_order=True)
        .agg(pl.col("val").is_not_null().all())
        .collect()
    )

    assert_frame_equal(expected_df, result_df)

    # edge case of empty series
    lf = pl.LazyFrame({"val": []}, schema={"val": pl.Int32})

    expected_df = pl.DataFrame({"val": [True]})
    result_df = lf.select(pl.col("val").is_not_null().all()).collect()
    assert_frame_equal(expected_df, result_df)


def test_is_not_null_followed_by_any() -> None:
    lf = pl.LazyFrame({"group": [0, 0, 0, 1, 2], "val": [6, 0, None, None, 5]})

    expected_df = pl.DataFrame({"group": [0, 1, 2], "val": [True, False, True]})
    result_lf = lf.group_by("group", maintain_order=True).agg(
        pl.col("val").is_not_null().any()
    )

    assert r'[[(col("val").null_count()) < (col("val").len())]]' in result_lf.explain()
    assert "is_not_null" not in result_lf.explain()
    assert_frame_equal(expected_df, result_lf.collect())

    # verify we don't optimize on chained expressions when last one is not col
    non_optimized_result_plan = (
        lf.group_by("group", maintain_order=True)
        .agg(pl.col("val").abs().is_not_null().any())
        .explain()
    )
    assert "null_count" not in non_optimized_result_plan
    assert "is_not_null" in non_optimized_result_plan

    # edge case of empty series
    lf = pl.LazyFrame({"val": []}, schema={"val": pl.Int32})

    expected_df = pl.DataFrame({"val": [False]})
    result_df = lf.select(pl.col("val").is_not_null().any()).collect()
    assert_frame_equal(expected_df, result_df)


def test_is_null_followed_by_sum() -> None:
    lf = pl.LazyFrame({"group": [0, 0, 0, 1, 2], "val": [6, 0, None, None, 5]})

    expected_df = pl.DataFrame(
        {"group": [0, 1, 2], "val": [1, 1, 0]}, schema_overrides={"val": pl.UInt32}
    )
    result_lf = lf.group_by("group", maintain_order=True).agg(
        pl.col("val").is_null().sum()
    )

    assert r'[col("val").null_count()]' in result_lf.explain()
    assert "is_null" not in result_lf.explain()
    assert_frame_equal(expected_df, result_lf.collect())

    # edge case of empty series
    lf = pl.LazyFrame({"val": []}, schema={"val": pl.Int32})

    expected_df = pl.DataFrame({"val": [0]}, schema={"val": pl.UInt32})
    result_df = lf.select(pl.col("val").is_null().sum()).collect()
    assert_frame_equal(expected_df, result_df)


def test_is_not_null_followed_by_sum() -> None:
    lf = pl.LazyFrame({"group": [0, 0, 0, 1, 2], "val": [6, 0, None, None, 5]})

    expected_df = pl.DataFrame(
        {"group": [0, 1, 2], "val": [2, 0, 1]}, schema_overrides={"val": pl.UInt32}
    )
    result_lf = lf.group_by("group", maintain_order=True).agg(
        pl.col("val").is_not_null().sum()
    )

    assert r'[[(col("val").len()) - (col("val").null_count())]]' in result_lf.explain()
    assert "is_not_null" not in result_lf.explain()
    assert_frame_equal(expected_df, result_lf.collect())

    # verify we don't optimize on chained expressions when last one is not col
    non_optimized_result_lf = lf.group_by("group", maintain_order=True).agg(
        pl.col("val").abs().is_not_null().sum()
    )
    assert "null_count" not in non_optimized_result_lf.explain()
    assert "is_not_null" in non_optimized_result_lf.explain()

    # edge case of empty series
    lf = pl.LazyFrame({"val": []}, schema={"val": pl.Int32})

    expected_df = pl.DataFrame({"val": [0]}, schema={"val": pl.UInt32})
    result_df = lf.select(pl.col("val").is_not_null().sum()).collect()
    assert_frame_equal(expected_df, result_df)


def test_drop_nulls_followed_by_len() -> None:
    lf = pl.LazyFrame({"group": [0, 0, 0, 1, 2], "val": [6, 0, None, None, 5]})

    expected_df = pl.DataFrame(
        {"group": [0, 1, 2], "val": [2, 0, 1]}, schema_overrides={"val": pl.UInt32}
    )
    result_lf = lf.group_by("group", maintain_order=True).agg(
        pl.col("val").drop_nulls().len()
    )

    assert r'[[(col("val").len()) - (col("val").null_count())]]' in result_lf.explain()
    assert "drop_nulls" not in result_lf.explain()
    assert_frame_equal(expected_df, result_lf.collect())

    # verify we don't optimize on chained expressions when last one is not col
    non_optimized_result_plan = (
        lf.group_by("group", maintain_order=True)
        .agg(pl.col("val").abs().drop_nulls().len())
        .explain()
    )
    assert "null_count" not in non_optimized_result_plan
    assert "drop_nulls" in non_optimized_result_plan


def test_drop_nulls_followed_by_count() -> None:
    lf = pl.LazyFrame({"group": [0, 0, 0, 1, 2], "val": [6, 0, None, None, 5]})

    expected_df = pl.DataFrame(
        {"group": [0, 1, 2], "val": [2, 0, 1]}, schema_overrides={"val": pl.UInt32}
    )
    result_lf = lf.group_by("group", maintain_order=True).agg(
        pl.col("val").drop_nulls().count()
    )

    assert r'[[(col("val").len()) - (col("val").null_count())]]' in result_lf.explain()
    assert "drop_nulls" not in result_lf.explain()
    assert_frame_equal(expected_df, result_lf.collect())

    # verify we don't optimize on chained expressions when last one is not col
    non_optimized_result_plan = (
        lf.group_by("group", maintain_order=True)
        .agg(pl.col("val").abs().drop_nulls().count())
        .explain()
    )
    assert "null_count" not in non_optimized_result_plan
    assert "drop_nulls" in non_optimized_result_plan


def test_collapse_joins() -> None:
    a = pl.LazyFrame({"a": [1, 2, 3], "b": [2, 2, 2]})
    b = pl.LazyFrame({"x": [7, 1, 2]})

    cross = a.join(b, how="cross")

    inner_join = cross.filter(pl.col.a == pl.col.x)
    e = inner_join.explain()
    assert "INNER JOIN" in e
    assert "FILTER" not in e
    assert_frame_equal(
        inner_join.collect(optimizations=pl.QueryOptFlags(collapse_joins=False)),
        inner_join.collect(),
        check_row_order=False,
    )

    inner_join = cross.filter(pl.col.x == pl.col.a)
    e = inner_join.explain()
    assert "INNER JOIN" in e
    assert "FILTER" not in e
    assert_frame_equal(
        inner_join.collect(optimizations=pl.QueryOptFlags(collapse_joins=False)),
        inner_join.collect(),
        check_row_order=False,
    )

    double_inner_join = cross.filter(pl.col.x == pl.col.a).filter(pl.col.x == pl.col.b)
    e = double_inner_join.explain()
    assert "INNER JOIN" in e
    assert "FILTER" not in e
    assert_frame_equal(
        double_inner_join.collect(optimizations=pl.QueryOptFlags(collapse_joins=False)),
        double_inner_join.collect(),
        check_row_order=False,
    )

    dont_mix = cross.filter(pl.col.x + pl.col.a != 0)
    e = dont_mix.explain()
    assert "NESTED LOOP JOIN" in e
    assert "FILTER" not in e
    assert_frame_equal(
        dont_mix.collect(optimizations=pl.QueryOptFlags(collapse_joins=False)),
        dont_mix.collect(),
        check_row_order=False,
    )

    iejoin = cross.filter(pl.col.x >= pl.col.a)
    e = iejoin.explain()
    assert "IEJOIN" in e
    assert "NESTED LOOP JOIN" not in e
    assert "CROSS JOIN" not in e
    assert "FILTER" not in e
    assert_frame_equal(
        iejoin.collect(optimizations=pl.QueryOptFlags(collapse_joins=False)),
        iejoin.collect(),
        check_row_order=False,
    )

    iejoin = cross.filter(pl.col.x >= pl.col.a).filter(pl.col.x <= pl.col.b)
    e = iejoin.explain()
    assert "IEJOIN" in e
    assert "CROSS JOIN" not in e
    assert "NESTED LOOP JOIN" not in e
    assert "FILTER" not in e
    assert_frame_equal(
        iejoin.collect(optimizations=pl.QueryOptFlags(collapse_joins=False)),
        iejoin.collect(),
        check_row_order=False,
    )


@pytest.mark.slow
def test_collapse_joins_combinations() -> None:
    # This just tests all possible combinations for expressions on a cross join.

    a = pl.LazyFrame({"a": [1, 2, 3], "x": [7, 2, 1]})
    b = pl.LazyFrame({"b": [2, 2, 2], "x": [7, 1, 3]})

    cross = a.join(b, how="cross")

    exprs = []

    for lhs in [pl.col.a, pl.col.b, pl.col.x, pl.lit(1), pl.col.a + pl.col.b]:
        for rhs in [pl.col.a, pl.col.b, pl.col.x, pl.lit(1), pl.col.a * pl.col.x]:
            for cmp in ["__eq__", "__ge__", "__lt__"]:
                e = (getattr(lhs, cmp))(rhs)
                exprs.append(e)

    for amount in range(3):
        for merge in itertools.product(["__and__", "__or__"] * (amount - 1)):
            for es in itertools.product(*([exprs] * amount)):
                e = es[0]
                for i in range(amount - 1):
                    e = (getattr(e, merge[i]))(es[i + 1])

                # NOTE: We need to sort because the order of the cross-join &
                # IE-join is unspecified. Therefore, this might not necessarily
                # create the exact same dataframe.
                optimized = cross.filter(e).sort(pl.all()).collect()
                unoptimized = cross.filter(e).collect(
                    optimizations=pl.QueryOptFlags(collapse_joins=False)
                )

                try:
                    assert_frame_equal(optimized, unoptimized, check_row_order=False)
                except:
                    print(e)
                    print()
                    print("Optimized")
                    print(cross.filter(e).explain())
                    print(optimized)
                    print()
                    print("Unoptimized")
                    print(
                        cross.filter(e).explain(
                            optimizations=pl.QueryOptFlags(collapse_joins=False)
                        )
                    )
                    print(unoptimized)
                    print()

                    raise


def test_order_observe_sort_before_unique_22485() -> None:
    lf = pl.LazyFrame(
        {
            "order": [3, 2, 1],
            "id": ["A", "A", "B"],
        }
    )

    expect = pl.DataFrame({"order": [1, 3], "id": ["B", "A"]})

    q = lf.sort("order").unique(["id"], keep="last").sort("order")

    plan = q.explain()
    assert "SORT BY" in plan[plan.index("UNIQUE") :]

    assert_frame_equal(q.collect(), expect)

    q = lf.sort("order").unique(["id"], keep="last", maintain_order=True)

    plan = q.explain()
    assert "SORT BY" in plan[plan.index("UNIQUE") :]

    assert_frame_equal(q.collect(), expect)


def test_order_observe_group_by() -> None:
    q = (
        pl.LazyFrame({"a": range(5)})
        .group_by("a", maintain_order=True)
        .agg(b=1)
        .sort("b")
    )

    plan = q.explain()
    assert "AGGREGATE[maintain_order: false]" in plan

    q = (
        pl.LazyFrame({"a": range(5)})
        .group_by("a", maintain_order=True)
        .agg(b=1)
        .sort("b", maintain_order=True)
    )

    plan = q.explain()
    assert "AGGREGATE[maintain_order: true]" in plan


def test_fused_correct_name() -> None:
    df = pl.DataFrame({"x": [1, 2, 3]})

    lf = df.lazy().select(
        (pl.col.x.alias("a") * pl.col.x.alias("b")) + pl.col.x.alias("c")
    )

    no_opts = lf.collect(optimizations=pl.QueryOptFlags.none())
    opts = lf.collect()
    assert_frame_equal(
        no_opts,
        opts,
    )
    assert_frame_equal(opts, pl.DataFrame({"a": [2, 6, 12]}))



from __future__ import annotations

import time
from datetime import date
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
import pytest

import polars as pl
from polars.exceptions import PolarsInefficientMapWarning
from polars.testing import assert_frame_equal

if TYPE_CHECKING:
    from polars._typing import JoinStrategy

pytestmark = pytest.mark.xdist_group("streaming")


def test_streaming_categoricals_5921() -> None:
    out_lazy = (
        pl.DataFrame({"X": ["a", "a", "a", "b", "b"], "Y": [2, 2, 2, 1, 1]})
        .lazy()
        .with_columns(pl.col("X").cast(pl.Categorical))
        .group_by("X")
        .agg(pl.col("Y").min())
        .sort("Y", descending=True)
        .collect(engine="streaming")
    )

    out_eager = (
        pl.DataFrame({"X": ["a", "a", "a", "b", "b"], "Y": [2, 2, 2, 1, 1]})
        .with_columns(pl.col("X").cast(pl.Categorical))
        .group_by("X")
        .agg(pl.col("Y").min())
        .sort("Y", descending=True)
    )

    for out in [out_eager, out_lazy]:
        assert out.dtypes == [pl.Categorical, pl.Int64]
        assert out.to_dict(as_series=False) == {"X": ["a", "b"], "Y": [2, 1]}


def test_streaming_block_on_literals_6054() -> None:
    df = pl.DataFrame({"col_1": [0] * 5 + [1] * 5})
    s = pl.Series("col_2", list(range(10)))

    assert df.lazy().with_columns(s).group_by("col_1").agg(pl.all().first()).collect(
        engine="streaming"
    ).sort("col_1").to_dict(as_series=False) == {"col_1": [0, 1], "col_2": [0, 5]}


@pytest.mark.may_fail_auto_streaming
def test_streaming_streamable_functions(monkeypatch: Any, capfd: Any) -> None:
    monkeypatch.setenv("POLARS_IDEAL_MORSEL_SIZE", "1")
    calls = 0

    def func(df: pl.DataFrame) -> pl.DataFrame:
        nonlocal calls
        calls += 1
        return df.with_columns(pl.col("a").alias("b"))

    assert (
        pl.DataFrame({"a": list(range(100))})
        .lazy()
        .map_batches(
            function=func,
            schema={"a": pl.Int64, "b": pl.Int64},
            streamable=True,
        )
    ).collect(engine="streaming").to_dict(as_series=False) == {
        "a": list(range(100)),
        "b": list(range(100)),
    }

    assert calls > 1


@pytest.mark.slow
@pytest.mark.may_fail_auto_streaming
def test_cross_join_stack() -> None:
    a = pl.Series(np.arange(100_000)).to_frame().lazy()
    t0 = time.time()
    assert a.join(a, how="cross").head().collect(engine="streaming").shape == (5, 2)
    t1 = time.time()
    assert (t1 - t0) < 0.5


def test_streaming_literal_expansion() -> None:
    df = pl.DataFrame(
        {
            "y": ["a", "b"],
            "z": [1, 2],
        }
    )

    q = df.lazy().select(
        x=pl.lit("constant"),
        y=pl.col("y"),
        z=pl.col("z"),
    )

    assert q.collect(engine="streaming").to_dict(as_series=False) == {
        "x": ["constant", "constant"],
        "y": ["a", "b"],
        "z": [1, 2],
    }
    assert q.group_by(["x", "y"]).agg(pl.mean("z")).sort("y").collect(
        engine="streaming"
    ).to_dict(as_series=False) == {
        "x": ["constant", "constant"],
        "y": ["a", "b"],
        "z": [1.0, 2.0],
    }
    assert q.group_by(["x"]).agg(pl.mean("z")).collect().to_dict(as_series=False) == {
        "x": ["constant"],
        "z": [1.5],
    }


@pytest.mark.may_fail_auto_streaming
def test_streaming_apply(monkeypatch: Any, capfd: Any) -> None:
    monkeypatch.setenv("POLARS_VERBOSE", "1")

    q = pl.DataFrame({"a": [1, 2]}).lazy()
    with pytest.warns(PolarsInefficientMapWarning, match="with this one instead"):
        (
            q.select(
                pl.col("a").map_elements(lambda x: x * 2, return_dtype=pl.Int64)
            ).collect(engine="streaming")
        )


def test_streaming_sortedness_propagation_9494() -> None:
    assert (
        pl.DataFrame(
            {
                "when": [date(2023, 5, 10), date(2023, 5, 20), date(2023, 6, 10)],
                "what": [1, 2, 3],
            }
        )
        .lazy()
        .sort("when")
        .group_by_dynamic("when", every="1mo")
        .agg(pl.col("what").sum())
        .collect(engine="streaming")
    ).to_dict(as_series=False) == {
        "when": [date(2023, 5, 1), date(2023, 6, 1)],
        "what": [3, 3],
    }


@pytest.mark.write_disk
@pytest.mark.slow
def test_streaming_generic_left_and_inner_join_from_disk(tmp_path: Path) -> None:
    tmp_path.mkdir(exist_ok=True)
    p0 = tmp_path / "df0.parquet"
    p1 = tmp_path / "df1.parquet"
    # by loading from disk, we get different chunks
    n = 200_000
    k = 100

    d0: dict[str, np.ndarray[Any, Any]] = {
        f"x{i}": np.random.random(n) for i in range(k)
    }
    d0.update({"id": np.arange(n)})

    df0 = pl.DataFrame(d0)
    df1 = df0.clone().select(pl.all().shuffle(111))

    df0.write_parquet(p0)
    df1.write_parquet(p1)

    lf0 = pl.scan_parquet(p0)
    lf1 = pl.scan_parquet(p1).select(pl.all().name.suffix("_r"))

    join_strategies: list[JoinStrategy] = ["left", "inner"]
    for how in join_strategies:
        assert_frame_equal(
            lf0.join(
                lf1, left_on="id", right_on="id_r", how=how, maintain_order="left"
            ).collect(engine="streaming"),
            lf0.join(lf1, left_on="id", right_on="id_r", how=how).collect(
                engine="in-memory"
            ),
            check_row_order=how == "left",
        )


def test_streaming_9776() -> None:
    df = pl.DataFrame({"col_1": ["a"] * 1000, "ID": [None] + ["a"] * 999})
    ordered = (
        df.group_by("col_1", "ID", maintain_order=True)
        .len()
        .filter(pl.col("col_1") == "a")
    )
    unordered = (
        df.group_by("col_1", "ID", maintain_order=False)
        .len()
        .filter(pl.col("col_1") == "a")
    )
    expected = [("a", None, 1), ("a", "a", 999)]
    assert ordered.rows() == expected
    assert unordered.sort(["col_1", "ID"]).rows() == expected


@pytest.mark.write_disk
def test_stream_empty_file(tmp_path: Path) -> None:
    p = tmp_path / "in.parquet"
    schema = {
        "KLN_NR": pl.String,
    }

    df = pl.DataFrame(
        {
            "KLN_NR": [],
        },
        schema=schema,
    )
    df.write_parquet(p)
    assert pl.scan_parquet(p).collect(engine="streaming").schema == schema


def test_streaming_empty_df() -> None:
    df = pl.DataFrame(
        [
            pl.Series("a", ["a", "b", "c", "b", "a", "a"], dtype=pl.Categorical()),
            pl.Series("b", ["b", "c", "c", "b", "a", "c"], dtype=pl.Categorical()),
        ]
    )

    result = (
        df.lazy()
        .join(df.lazy(), on="a", how="inner")
        .filter(False)
        .collect(engine="streaming")
    )

    assert result.to_dict(as_series=False) == {"a": [], "b": [], "b_right": []}


def test_streaming_duplicate_cols_5537() -> None:
    assert pl.DataFrame({"a": [1, 2, 3], "b": [1, 2, 3]}).lazy().with_columns(
        (pl.col("a") * 2).alias("foo"), (pl.col("a") * 3)
    ).collect(engine="streaming").to_dict(as_series=False) == {
        "a": [3, 6, 9],
        "b": [1, 2, 3],
        "foo": [2, 4, 6],
    }


def test_null_sum_streaming_10455() -> None:
    df = pl.DataFrame(
        {
            "x": [1] * 10,
            "y": [None] * 10,
        },
        schema={"x": pl.Int64, "y": pl.Float32},
    )
    result = df.lazy().group_by("x").sum().collect(engine="streaming")
    expected = {"x": [1], "y": [0.0]}
    assert result.to_dict(as_series=False) == expected


def test_boolean_agg_schema() -> None:
    df = pl.DataFrame(
        {
            "x": [1, 1, 1],
            "y": [False, True, False],
        }
    ).lazy()

    agg_df = df.group_by("x").agg(pl.col("y").max().alias("max_y"))

    for streaming in [True, False]:
        assert (
            agg_df.collect(engine="streaming" if streaming else "in-memory").schema
            == agg_df.collect_schema()
            == {"x": pl.Int64, "max_y": pl.Boolean}
        )


@pytest.mark.write_disk
def test_streaming_csv_headers_but_no_data_13770(tmp_path: Path) -> None:
    with Path.open(tmp_path / "header_no_data.csv", "w") as f:
        f.write("name, age\n")

    schema = {"name": pl.String, "age": pl.Int32}
    df = (
        pl.scan_csv(tmp_path / "header_no_data.csv", schema=schema)
        .head()
        .collect(engine="streaming")
    )
    assert df.height == 0
    assert df.schema == schema


@pytest.mark.write_disk
def test_streaming_with_hconcat(tmp_path: Path) -> None:
    df1 = pl.DataFrame(
        {
            "id": [0, 0, 1, 1, 2, 2],
            "x": [0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
        }
    )
    df1.write_parquet(tmp_path / "df1.parquet")

    df2 = pl.DataFrame(
        {
            "y": [6.0, 7.0, 8.0, 9.0, 10.0, 11.0],
        }
    )
    df2.write_parquet(tmp_path / "df2.parquet")

    lf1 = pl.scan_parquet(tmp_path / "df1.parquet")
    lf2 = pl.scan_parquet(tmp_path / "df2.parquet")
    query = (
        pl.concat([lf1, lf2], how="horizontal")
        .group_by("id")
        .agg(pl.all().mean())
        .sort(pl.col("id"))
    )

    result = query.collect(engine="streaming")

    expected = pl.DataFrame(
        {
            "id": [0, 1, 2],
            "x": [0.5, 2.5, 4.5],
            "y": [6.5, 8.5, 10.5],
        }
    )

    assert_frame_equal(result, expected)


@pytest.mark.write_disk
def test_elementwise_identification_in_ternary_15767(tmp_path: Path) -> None:
    tmp_path.mkdir(exist_ok=True)

    (
        pl.LazyFrame({"a": pl.Series([1])})
        .with_columns(b=pl.col("a").is_in(pl.Series([1, 2, 3])))
        .sink_parquet(tmp_path / "1")
    )

    (
        pl.LazyFrame({"a": pl.Series([1])})
        .with_columns(
            b=pl.when(pl.col("a").is_in(pl.Series([1, 2, 3]))).then(pl.col("a"))
        )
        .sink_parquet(tmp_path / "1")
    )


def test_streaming_temporal_17669() -> None:
    df = (
        pl.LazyFrame({"a": [1, 2, 3]}, schema={"a": pl.Datetime("us")})
        .with_columns(
            b=pl.col("a").dt.date(),
            c=pl.col("a").dt.time(),
        )
        .collect(engine="streaming")
    )
    assert df.schema == {
        "a": pl.Datetime("us"),
        "b": pl.Date,
        "c": pl.Time,
    }


def test_i128_sum_reduction() -> None:
    assert (
        pl.Series("a", [1, 2, 3], pl.Int128)
        .to_frame()
        .lazy()
        .sum()
        .collect(engine="streaming")
        .item()
        == 6
    )



from __future__ import annotations

from datetime import date, datetime, time, timedelta
from typing import TYPE_CHECKING, Any

import pytest

import polars as pl
from polars.exceptions import ShapeError
from polars.testing import assert_frame_equal, assert_series_equal

if TYPE_CHECKING:
    from polars._typing import PolarsDataType


@pytest.mark.parametrize(
    ("const", "dtype"),
    [
        (1, pl.Int8),
        (4, pl.UInt32),
        (4.5, pl.Float32),
        (None, pl.Float64),
        ("白鵬翔", pl.String),
        (date.today(), pl.Date),
        (datetime.now(), pl.Datetime("ns")),
        (time(23, 59, 59), pl.Time),
        (timedelta(hours=7, seconds=123), pl.Duration("ms")),
    ],
)
def test_extend_constant(const: Any, dtype: PolarsDataType) -> None:
    df = pl.DataFrame({"a": pl.Series("s", [None], dtype=dtype)})

    expected_df = pl.DataFrame(
        {"a": pl.Series("s", [None, const, const, const], dtype=dtype)}
    )

    assert_frame_equal(df.select(pl.col("a").extend_constant(const, 3)), expected_df)

    s = pl.Series("s", [None], dtype=dtype)
    expected = pl.Series("s", [None, const, const, const], dtype=dtype)
    assert_series_equal(s.extend_constant(const, 3), expected)

    # test n expr
    expected = pl.Series("s", [None, const, const], dtype=dtype)
    assert_series_equal(s.extend_constant(const, pl.lit(2)), expected)

    # test value expr
    expected = pl.Series("s", [None, const, const, const], dtype=dtype)
    assert_series_equal(s.extend_constant(pl.lit(const, dtype=dtype), 3), expected)


@pytest.mark.parametrize(
    ("const", "dtype"),
    [
        (1, pl.Int8),
        (4, pl.UInt32),
        (4.5, pl.Float32),
        (None, pl.Float64),
        ("白鵬翔", pl.String),
        (date.today(), pl.Date),
        (datetime.now(), pl.Datetime("ns")),
        (time(23, 59, 59), pl.Time),
        (timedelta(hours=7, seconds=123), pl.Duration("ms")),
    ],
)
def test_extend_constant_arr(const: Any, dtype: PolarsDataType) -> None:
    """
    Test extend_constant in pl.List array.

    NOTE: This function currently fails when the Series is a list with a single [None]
          value. Hence, this function does not begin with [[None]], but [[const]].
    """
    s = pl.Series("s", [[const]], dtype=pl.List(dtype))

    expected = pl.Series("s", [[const, const, const, const]], dtype=pl.List(dtype))

    assert_series_equal(s.list.eval(pl.element().extend_constant(const, 3)), expected)


def test_extend_by_not_uint_expr() -> None:
    s = pl.Series("s", [1])
    with pytest.raises(ShapeError, match="'value' must be a scalar value"):
        s.extend_constant(pl.Series([2, 3]), 3)
    with pytest.raises(ShapeError, match="'n' must be a scalar value"):
        s.extend_constant(2, pl.Series([3, 4]))



import io

import polars as pl
from polars.testing import assert_frame_equal


def test_fast_count_alias_18581() -> None:
    f = io.BytesIO()
    f.write(b"a,b,c\n1,2,3\n4,5,6")
    f.flush()
    f.seek(0)

    df = pl.scan_csv(f).select(pl.len().alias("weird_name")).collect()

    assert_frame_equal(pl.DataFrame({"weird_name": 2}), df)



from __future__ import annotations

import io
from typing import IO, TYPE_CHECKING, Any, Callable

import pyarrow.parquet as pq
import pytest
from hypothesis import given
from hypothesis import strategies as st

import polars as pl
from polars.meta.index_type import get_index_type
from polars.testing import assert_frame_equal

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.write_disk
@pytest.mark.parametrize(
    ("scan", "write"),
    [
        (pl.scan_ipc, pl.DataFrame.write_ipc),
        (pl.scan_parquet, pl.DataFrame.write_parquet),
        (pl.scan_csv, pl.DataFrame.write_csv),
        (pl.scan_ndjson, pl.DataFrame.write_ndjson),
    ],
)
def test_include_file_paths(tmp_path: Path, scan: Any, write: Any) -> None:
    a_path = tmp_path / "a"
    b_path = tmp_path / "b"

    write(pl.DataFrame({"a": [5, 10]}), a_path)
    write(pl.DataFrame({"a": [1996]}), b_path)

    out = scan([a_path, b_path], include_file_paths="f")

    assert_frame_equal(
        out.collect(),
        pl.DataFrame(
            {
                "a": [5, 10, 1996],
                "f": [str(a_path), str(a_path), str(b_path)],
            }
        ),
    )


@pytest.mark.parametrize(
    ("scan", "write", "ext", "supports_missing_columns", "supports_hive_partitioning"),
    [
        (pl.scan_ipc, pl.DataFrame.write_ipc, "ipc", False, True),
        (pl.scan_parquet, pl.DataFrame.write_parquet, "parquet", True, True),
        (pl.scan_csv, pl.DataFrame.write_csv, "csv", False, False),
        (pl.scan_ndjson, pl.DataFrame.write_ndjson, "jsonl", False, False),
    ],
)
@pytest.mark.parametrize("missing_column", [False, True])
@pytest.mark.parametrize("row_index", [False, True])
@pytest.mark.parametrize("include_file_paths", [False, True])
@pytest.mark.parametrize("hive", [False, True])
@pytest.mark.parametrize("col", [False, True])
@pytest.mark.write_disk
def test_multiscan_projection(
    tmp_path: Path,
    scan: Callable[..., pl.LazyFrame],
    write: Callable[[pl.DataFrame, Path], Any],
    ext: str,
    supports_missing_columns: bool,
    supports_hive_partitioning: bool,
    missing_column: bool,
    row_index: bool,
    include_file_paths: bool,
    hive: bool,
    col: bool,
) -> None:
    a = pl.DataFrame({"col": [5, 10, 1996]})
    b = pl.DataFrame({"col": [13, 37]})

    if missing_column and supports_missing_columns:
        a = a.with_columns(missing=pl.Series([420, 2000, 9]))

    a_path: Path
    b_path: Path
    multiscan_path: Path

    if hive and supports_hive_partitioning:
        (tmp_path / "hive_col=0").mkdir()
        a_path = tmp_path / "hive_col=0" / f"a.{ext}"
        (tmp_path / "hive_col=1").mkdir()
        b_path = tmp_path / "hive_col=1" / f"b.{ext}"

        multiscan_path = tmp_path

    else:
        a_path = tmp_path / f"a.{ext}"
        b_path = tmp_path / f"b.{ext}"

        multiscan_path = tmp_path / f"*.{ext}"

    write(a, a_path)
    write(b, b_path)

    base_projection = []
    if missing_column and supports_missing_columns:
        base_projection += ["missing"]
    if row_index:
        base_projection += ["row_index"]
    if include_file_paths:
        base_projection += ["file_path"]
    if hive and supports_hive_partitioning:
        base_projection += ["hive_col"]
    if col:
        base_projection += ["col"]

    ifp = "file_path" if include_file_paths else None
    ri = "row_index" if row_index else None

    args = {
        "missing_columns": "insert" if missing_column else "raise",
        "include_file_paths": ifp,
        "row_index_name": ri,
        "hive_partitioning": hive,
    }

    if not supports_missing_columns:
        del args["missing_columns"]
    if not supports_hive_partitioning:
        del args["hive_partitioning"]

    for projection in [
        base_projection,
        base_projection[::-1],
    ]:
        assert_frame_equal(
            scan(multiscan_path, **args).collect(engine="streaming").select(projection),
            scan(multiscan_path, **args).select(projection).collect(engine="streaming"),
        )

    for remove in range(len(base_projection)):
        new_projection = base_projection.copy()
        new_projection.pop(remove)

        for projection in [
            new_projection,
            new_projection[::-1],
        ]:
            print(projection)
            assert_frame_equal(
                scan(multiscan_path, **args)
                .collect(engine="streaming")
                .select(projection),
                scan(multiscan_path, **args)
                .select(projection)
                .collect(engine="streaming"),
            )


@pytest.mark.parametrize(
    ("scan", "write", "ext"),
    [
        (pl.scan_ipc, pl.DataFrame.write_ipc, "ipc"),
        (pl.scan_parquet, pl.DataFrame.write_parquet, "parquet"),
    ],
)
@pytest.mark.write_disk
def test_multiscan_hive_predicate(
    tmp_path: Path,
    scan: Callable[..., pl.LazyFrame],
    write: Callable[[pl.DataFrame, Path], Any],
    ext: str,
) -> None:
    a = pl.DataFrame({"col": [5, 10, 1996]})
    b = pl.DataFrame({"col": [13, 37]})
    c = pl.DataFrame({"col": [3, 5, 2024]})

    (tmp_path / "hive_col=0").mkdir()
    a_path = tmp_path / "hive_col=0" / f"0.{ext}"
    (tmp_path / "hive_col=1").mkdir()
    b_path = tmp_path / "hive_col=1" / f"0.{ext}"
    (tmp_path / "hive_col=2").mkdir()
    c_path = tmp_path / "hive_col=2" / f"0.{ext}"

    multiscan_path = tmp_path

    write(a, a_path)
    write(b, b_path)
    write(c, c_path)

    full = scan(multiscan_path).collect(engine="streaming")
    full_ri = full.with_row_index("ri", 42)

    last_pred = None
    try:
        for pred in [
            pl.col.hive_col == 0,
            pl.col.hive_col == 1,
            pl.col.hive_col == 2,
            pl.col.hive_col < 2,
            pl.col.hive_col > 0,
            pl.col.hive_col != 1,
            pl.col.hive_col != 3,
            pl.col.col == 13,
            pl.col.col != 13,
            (pl.col.col != 13) & (pl.col.hive_col == 1),
            (pl.col.col != 13) & (pl.col.hive_col != 1),
        ]:
            last_pred = pred
            assert_frame_equal(
                full.filter(pred),
                scan(multiscan_path).filter(pred).collect(engine="streaming"),
            )

            assert_frame_equal(
                full_ri.filter(pred),
                scan(multiscan_path)
                .with_row_index("ri", 42)
                .filter(pred)
                .collect(engine="streaming"),
            )
    except Exception as _:
        print(last_pred)
        raise


@pytest.mark.parametrize(
    ("scan", "write", "ext"),
    [
        (pl.scan_ipc, pl.DataFrame.write_ipc, "ipc"),
        (pl.scan_parquet, pl.DataFrame.write_parquet, "parquet"),
        (pl.scan_csv, pl.DataFrame.write_csv, "csv"),
        (pl.scan_ndjson, pl.DataFrame.write_ndjson, "jsonl"),
    ],
)
@pytest.mark.write_disk
def test_multiscan_row_index(
    tmp_path: Path,
    scan: Callable[..., pl.LazyFrame],
    write: Callable[[pl.DataFrame, Path], Any],
    ext: str,
) -> None:
    a = pl.DataFrame({"col": [5, 10, 1996]})
    b = pl.DataFrame({"col": [42]})
    c = pl.DataFrame({"col": [13, 37]})

    write(a, tmp_path / f"a.{ext}")
    write(b, tmp_path / f"b.{ext}")
    write(c, tmp_path / f"c.{ext}")

    col = pl.concat([a, b, c]).to_series()
    g = tmp_path / f"*.{ext}"

    assert_frame_equal(
        scan(g, row_index_name="ri").collect(),
        pl.DataFrame(
            [
                pl.Series("ri", range(6), get_index_type()),
                col,
            ]
        ),
    )

    start = 42
    assert_frame_equal(
        scan(g, row_index_name="ri", row_index_offset=start).collect(),
        pl.DataFrame(
            [
                pl.Series("ri", range(start, start + 6), get_index_type()),
                col,
            ]
        ),
    )

    start = 42
    assert_frame_equal(
        scan(g, row_index_name="ri", row_index_offset=start).slice(3, 3).collect(),
        pl.DataFrame(
            [
                pl.Series("ri", range(start + 3, start + 6), get_index_type()),
                col.slice(3, 3),
            ]
        ),
    )

    start = 42
    assert_frame_equal(
        scan(g, row_index_name="ri", row_index_offset=start)
        .filter(pl.col("col") < 15)
        .collect(),
        pl.DataFrame(
            [
                pl.Series("ri", [start + 0, start + 1, start + 4], get_index_type()),
                pl.Series("col", [5, 10, 13]),
            ]
        ),
    )


@pytest.mark.parametrize(
    ("scan", "write", "ext"),
    [
        (pl.scan_ipc, pl.DataFrame.write_ipc, "ipc"),
        (pl.scan_parquet, pl.DataFrame.write_parquet, "parquet"),
        pytest.param(
            pl.scan_csv,
            pl.DataFrame.write_csv,
            "csv",
            marks=pytest.mark.xfail(
                reason="See https://github.com/pola-rs/polars/issues/21211"
            ),
        ),
        (pl.scan_ndjson, pl.DataFrame.write_ndjson, "jsonl"),
    ],
)
@pytest.mark.write_disk
def test_schema_mismatch_type_mismatch(
    tmp_path: Path,
    scan: Callable[..., pl.LazyFrame],
    write: Callable[[pl.DataFrame, Path], Any],
    ext: str,
) -> None:
    a = pl.DataFrame({"xyz_col": [5, 10, 1996]})
    b = pl.DataFrame({"xyz_col": ["a", "b", "c"]})

    a_path = tmp_path / f"a.{ext}"
    b_path = tmp_path / f"b.{ext}"

    multiscan_path = tmp_path / f"*.{ext}"

    write(a, a_path)
    write(b, b_path)

    q = scan(multiscan_path)

    # NDJSON will just parse according to `projected_schema`
    cx = (
        pytest.raises(pl.exceptions.ComputeError, match="cannot parse 'a' as Int64")
        if scan is pl.scan_ndjson
        else pytest.raises(
            pl.exceptions.SchemaError,
            match=(
                "data type mismatch for column xyz_col: "
                "incoming: String != target: Int64"
            ),
        )
    )

    with cx:
        q.collect(engine="streaming")


@pytest.mark.parametrize(
    ("scan", "write", "ext"),
    [
        # (pl.scan_parquet, pl.DataFrame.write_parquet, "parquet"), # TODO: _
        # (pl.scan_ipc, pl.DataFrame.write_ipc, "ipc"), # TODO: _
        pytest.param(
            pl.scan_csv,
            pl.DataFrame.write_csv,
            "csv",
            marks=pytest.mark.xfail(
                reason="See https://github.com/pola-rs/polars/issues/21211"
            ),
        ),
        # (pl.scan_ndjson, pl.DataFrame.write_ndjson, "jsonl"), # TODO: _
    ],
)
@pytest.mark.write_disk
def test_schema_mismatch_order_mismatch(
    tmp_path: Path,
    scan: Callable[..., pl.LazyFrame],
    write: Callable[[pl.DataFrame, Path], Any],
    ext: str,
) -> None:
    a = pl.DataFrame({"x": [5, 10, 1996], "y": ["a", "b", "c"]})
    b = pl.DataFrame({"y": ["x", "y"], "x": [1, 2]})

    a_path = tmp_path / f"a.{ext}"
    b_path = tmp_path / f"b.{ext}"

    multiscan_path = tmp_path / f"*.{ext}"

    write(a, a_path)
    write(b, b_path)

    q = scan(multiscan_path)

    with pytest.raises(pl.exceptions.SchemaError):
        q.collect(engine="streaming")


@pytest.mark.parametrize(
    ("scan", "write"),
    [
        (pl.scan_ipc, pl.DataFrame.write_ipc),
        (pl.scan_parquet, pl.DataFrame.write_parquet),
        (
            pl.scan_csv,
            pl.DataFrame.write_csv,
        ),
        (pl.scan_ndjson, pl.DataFrame.write_ndjson),
    ],
)
def test_multiscan_head(
    scan: Callable[..., pl.LazyFrame],
    write: Callable[[pl.DataFrame, io.BytesIO | Path], Any],
) -> None:
    a = io.BytesIO()
    b = io.BytesIO()
    for f in [a, b]:
        write(pl.Series("c1", range(10)).to_frame(), f)
        f.seek(0)

    assert_frame_equal(
        scan([a, b]).head(5).collect(engine="streaming"),
        pl.Series("c1", range(5)).to_frame(),
    )


@pytest.mark.parametrize(
    ("scan", "write"),
    [
        (pl.scan_ipc, pl.DataFrame.write_ipc),
        (pl.scan_parquet, pl.DataFrame.write_parquet),
        (pl.scan_ndjson, pl.DataFrame.write_ndjson),
        (
            pl.scan_csv,
            pl.DataFrame.write_csv,
        ),
    ],
)
def test_multiscan_tail(
    scan: Callable[..., pl.LazyFrame],
    write: Callable[[pl.DataFrame, io.BytesIO | Path], Any],
) -> None:
    a = io.BytesIO()
    b = io.BytesIO()
    for f in [a, b]:
        write(pl.Series("c1", range(10)).to_frame(), f)
        f.seek(0)

    assert_frame_equal(
        scan([a, b]).tail(5).collect(engine="streaming"),
        pl.Series("c1", range(5, 10)).to_frame(),
    )


@pytest.mark.parametrize(
    ("scan", "write"),
    [
        (pl.scan_ipc, pl.DataFrame.write_ipc),
        (pl.scan_parquet, pl.DataFrame.write_parquet),
        (pl.scan_ndjson, pl.DataFrame.write_ndjson),
        (
            pl.scan_csv,
            pl.DataFrame.write_csv,
        ),
    ],
)
def test_multiscan_slice_middle(
    scan: Callable[..., pl.LazyFrame],
    write: Callable[[pl.DataFrame, io.BytesIO | Path], Any],
) -> None:
    fs = [io.BytesIO() for _ in range(13)]
    for f in fs:
        write(pl.Series("c1", range(7)).to_frame(), f)
        f.seek(0)

    offset = 5 * 7 - 5
    expected = (
        list(range(2, 7))  # fs[4]
        + list(range(7))  # fs[5]
        + list(range(5))  # fs[6]
    )
    expected_series = [pl.Series("c1", expected)]
    ri_expected_series = [
        pl.Series("ri", range(offset, offset + 17), get_index_type())
    ] + expected_series

    assert_frame_equal(
        scan(fs).slice(offset, 17).collect(engine="streaming"),
        pl.DataFrame(expected_series),
    )
    assert_frame_equal(
        scan(fs, row_index_name="ri").slice(offset, 17).collect(engine="streaming"),
        pl.DataFrame(ri_expected_series),
    )

    # Negative slices
    offset = -(13 * 7 - offset)
    assert_frame_equal(
        scan(fs).slice(offset, 17).collect(engine="streaming"),
        pl.DataFrame(expected_series),
    )
    assert_frame_equal(
        scan(fs, row_index_name="ri").slice(offset, 17).collect(engine="streaming"),
        pl.DataFrame(ri_expected_series),
    )


@pytest.mark.parametrize(
    ("scan", "write", "ext"),
    [
        (pl.scan_ipc, pl.DataFrame.write_ipc, "ipc"),
        (pl.scan_parquet, pl.DataFrame.write_parquet, "parquet"),
        (pl.scan_ndjson, pl.DataFrame.write_ndjson, "jsonl"),
        (pl.scan_csv, pl.DataFrame.write_csv, "csv"),
    ],
)
@given(offset=st.integers(-100, 100), length=st.integers(0, 101))
def test_multiscan_slice_parametric(
    scan: Callable[..., pl.LazyFrame],
    write: Callable[[pl.DataFrame, io.BytesIO | Path], Any],
    ext: str,
    offset: int,
    length: int,
) -> None:
    ref = io.BytesIO()
    write(pl.Series("c1", [i % 7 for i in range(13 * 7)]).to_frame(), ref)
    ref.seek(0)

    fs = [io.BytesIO() for _ in range(13)]
    for f in fs:
        write(pl.Series("c1", range(7)).to_frame(), f)
        f.seek(0)

    assert_frame_equal(
        scan(ref).slice(offset, length).collect(),
        scan(fs).slice(offset, length).collect(engine="streaming"),
    )

    ref.seek(0)
    for f in fs:
        f.seek(0)

    assert_frame_equal(
        scan(ref, row_index_name="ri", row_index_offset=42)
        .slice(offset, length)
        .collect(),
        scan(fs, row_index_name="ri", row_index_offset=42)
        .slice(offset, length)
        .collect(engine="streaming"),
    )

    assert_frame_equal(
        scan(ref, row_index_name="ri", row_index_offset=42)
        .slice(offset, length)
        .select("ri")
        .collect(),
        scan(fs, row_index_name="ri", row_index_offset=42)
        .slice(offset, length)
        .select("ri")
        .collect(engine="streaming"),
    )


@pytest.mark.parametrize(
    ("scan", "write"),
    [
        (pl.scan_ipc, pl.DataFrame.write_ipc),
        (pl.scan_parquet, pl.DataFrame.write_parquet),
        (pl.scan_csv, pl.DataFrame.write_csv),
        (pl.scan_ndjson, pl.DataFrame.write_ndjson),
    ],
)
def test_many_files(tmp_path: Path, scan: Any, write: Any) -> None:
    f = io.BytesIO()
    write(pl.DataFrame({"a": [5, 10, 1996]}), f)
    bs = f.getvalue()

    out = scan([bs] * 1023)

    assert_frame_equal(
        out.collect(),
        pl.DataFrame(
            {
                "a": [5, 10, 1996] * 1023,
            }
        ),
    )


def test_deadlock_stop_requested(tmp_path: Path, monkeypatch: Any) -> None:
    df = pl.DataFrame(
        {
            "a": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        }
    )

    f = io.BytesIO()
    df.write_parquet(f, row_group_size=1)

    monkeypatch.setenv("POLARS_MAX_THREADS", "2")
    monkeypatch.setenv("POLARS_JOIN_SAMPLE_LIMIT", "1")

    left_fs = [io.BytesIO(f.getbuffer()) for _ in range(10)]
    right_fs = [io.BytesIO(f.getbuffer()) for _ in range(10)]

    left = pl.scan_parquet(left_fs)  # type: ignore[arg-type]
    right = pl.scan_parquet(right_fs)  # type: ignore[arg-type]

    left.join(right, pl.col.a == pl.col.a).collect(engine="streaming").height


@pytest.mark.parametrize(
    ("scan", "write"),
    [
        (pl.scan_ipc, pl.DataFrame.write_ipc),
        (pl.scan_parquet, pl.DataFrame.write_parquet),
        (pl.scan_csv, pl.DataFrame.write_csv),
        (pl.scan_ndjson, pl.DataFrame.write_ndjson),
    ],
)
def test_deadlock_linearize(scan: Any, write: Any) -> None:
    df = pl.DataFrame(
        {
            "a": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        }
    )

    f = io.BytesIO()
    write(df, f)
    fs = [io.BytesIO(f.getbuffer()) for _ in range(10)]
    lf = scan(fs).head(100)

    assert_frame_equal(
        lf.collect(
            engine="streaming", optimizations=pl.QueryOptFlags(slice_pushdown=False)
        ),
        pl.concat([df] * 10),
    )


@pytest.mark.parametrize(
    ("scan", "write"),
    [
        (pl.scan_ipc, pl.DataFrame.write_ipc),
        (pl.scan_parquet, pl.DataFrame.write_parquet),
        (pl.scan_csv, pl.DataFrame.write_csv),
        (pl.scan_ndjson, pl.DataFrame.write_ndjson),
    ],
)
def test_row_index_filter_22612(scan: Any, write: Any) -> None:
    df = pl.DataFrame(
        {
            "a": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        }
    )

    f = io.BytesIO()

    if write is pl.DataFrame.write_parquet:
        df.write_parquet(f, row_group_size=5)
        assert pq.read_metadata(f).num_row_groups == 2
    else:
        write(df, f)

    for end in range(2, 10):
        assert_frame_equal(
            scan(f)
            .with_row_index()
            .filter(pl.col("index") >= end - 2, pl.col("index") <= end)
            .collect(),
            df.with_row_index().slice(end - 2, 3),
        )

        assert_frame_equal(
            scan(f)
            .with_row_index()
            .filter(pl.col("index").is_between(end - 2, end))
            .collect(),
            df.with_row_index().slice(end - 2, 3),
        )


@pytest.mark.parametrize(
    ("scan", "write"),
    [
        (pl.scan_ipc, pl.DataFrame.write_ipc),
        (pl.scan_parquet, pl.DataFrame.write_parquet),
        (pl.scan_csv, pl.DataFrame.write_csv),
        (pl.scan_ndjson, pl.DataFrame.write_ndjson),
    ],
)
def test_row_index_name_in_file(scan: Any, write: Any) -> None:
    f = io.BytesIO()
    write(pl.DataFrame({"index": 1}), f)

    with pytest.raises(
        pl.exceptions.DuplicateError,
        match="cannot add row_index with name 'index': column already exists in file",
    ):
        scan(f).with_row_index().collect()


def test_extra_columns_not_ignored_22218() -> None:
    dfs = [pl.DataFrame({"a": 1, "b": 1}), pl.DataFrame({"a": 2, "c": 2})]

    files: list[IO[bytes]] = [io.BytesIO(), io.BytesIO()]

    dfs[0].write_parquet(files[0])
    dfs[1].write_parquet(files[1])

    with pytest.raises(
        pl.exceptions.SchemaError,
        match="extra column in file outside of expected schema: c, hint: specify .*or pass",
    ):
        (pl.scan_parquet(files, missing_columns="insert").select(pl.all()).collect())

    assert_frame_equal(
        pl.scan_parquet(
            files,
            missing_columns="insert",
            extra_columns="ignore",
        )
        .select(pl.all())
        .collect(),
        pl.DataFrame({"a": [1, 2], "b": [1, None]}),
    )
